# -*- coding: utf-8 -*-
"""
  Implements higher mathematical functions.

  The functions are discribed in :doc:`Mellin space </Theory/Mellin>`.
"""

import numpy as np
import scipy.special
import numba as nb

# compute constants only once
zeta2 = scipy.special.zeta(2)
zeta3 = scipy.special.zeta(3)


@nb.njit("c16(c16,u1)", cache=True)
def cern_polygamma(Z, K: int):  # pylint: disable=all
    """
    Computes the polygamma functions :math:`\\psi_k(z)`.

    Reimplementation of ``WPSIPG`` (C317) in `CERNlib <http://cernlib.web.cern.ch/cernlib/>`_
    :cite:`KOLBIG1972221`.

    Note that the SciPy implementation :data:`scipy.special.digamma`
    does not allow for complex inputs.

    Parameters
    ----------
      Z : complex
        argument of polygamma function
      K : int
        order of polygamma function

    Returns
    -------
      H : complex
        k-th polygamma function :math:`\\psi_k(z)`
    """
    # fmt: off
    DELTA = 5e-13
    R1 = 1
    HF = R1/2
    C1 = np.pi**2
    C2 = 2*np.pi**3
    C3 = 2*np.pi**4
    C4 = 8*np.pi**5

    # SGN is originally indexed 0:4 -> no shift
    SGN = [-1,+1,-1,+1,-1]
    # FCT is originally indexed -1:4 -> shift +1
    FCT = [0,1,1,2,6,24]

    # C is originally indexed 1:6 x 0:4 -> swap indices and shift new last -1
    C = nb.typed.List()
    C.append([
            8.33333333333333333e-2,
           -8.33333333333333333e-3,
            3.96825396825396825e-3,
           -4.16666666666666667e-3,
            7.57575757575757576e-3,
           -2.10927960927960928e-2])
    C.append([
            1.66666666666666667e-1,
           -3.33333333333333333e-2,
            2.38095238095238095e-2,
           -3.33333333333333333e-2,
            7.57575757575757576e-2,
           -2.53113553113553114e-1])
    C.append([
            5.00000000000000000e-1,
           -1.66666666666666667e-1,
            1.66666666666666667e-1,
           -3.00000000000000000e-1,
            8.33333333333333333e-1,
           -3.29047619047619048e+0])
    C.append([
            2.00000000000000000e+0,
           -1.00000000000000000e+0,
            1.33333333333333333e+0,
           -3.00000000000000000e+0,
            1.00000000000000000e+1,
           -4.60666666666666667e+1])
    C.append([10., -7., 12., -33., 130., -691.])
    U=Z
    X=np.real(U)
    A=np.abs(X)
    if K < 0 or K > 4:
        raise NotImplementedError("Order K has to be in [0:4]")
    if np.abs(np.imag(U)) < DELTA and np.abs(X+int(A)) < DELTA:
        raise ValueError("Argument Z equals non-positive integer")
    K1=K+1
    if X < 0:
        U=-U
    V=U
    H=0
    if A < 15:
        H=1/V**K1
        for I in range(1,14-int(A)+1):
            V=V+1
            H=H+1/V**K1
        V=V+1
    R=1/V**2
    P=R*C[K][6-1]
    for I in range(5,1-1,-1):
        P=R*(C[K][I-1]+P)
    H=SGN[K]*(FCT[K+1]*H+(V*(FCT[K-1+1]+P)+HF*FCT[K+1])/V**K1)
    if K == 0:
        H=H+np.log(V)
    if X < 0:
        V=np.pi*U
        X=np.real(V)
        Y=np.imag(V)
        A=np.sin(X)
        B=np.cos(X)
        T=np.tanh(Y)
        P=np.complex(B,-A*T)/np.complex(A,B*T)
        if K == 0:
            H=H+1/U+np.pi*P
        elif K == 1:
            H=-H+1/U**2+C1*(P**2+1)
        elif K == 2:
            H=H+2/U**3+C2*P*(P**2+1)
        elif K == 3:
            R=P**2
            H=-H+6/U**4+C3*((3*R+4)*R+1)
        elif K == 4:
            R=P**2
            H=H+24/U**5+C4*P*((3*R+5)*R+2)
    return H
    # fmt: on


@nb.njit("c16(c16)", cache=True)
def harmonic_S1(N):
    r"""
    Computes the harmonic sum :math:`S_1(N)`.

    .. math::
      S_1(N) = \sum\limits_{j=1}^N \frac 1 j = \psi_0(N+1)+\gamma_E

    with :math:`\psi_0(N)` the digamma function and :math:`\gamma_E` the
    Euler-Mascheroni constant.

    Parameters
    ----------
      N : complex
        Mellin moment

    Returns
    -------
      S_1 : complex
        (simple) Harmonic sum :math:`S_1(N)`

    See Also
    --------
      cern_polygamma : :math:`\psi_0(N)`
    """
    return cern_polygamma(N + 1.0, 0) + np.euler_gamma


@nb.njit("c16(c16)", cache=True)
def harmonic_S2(N):
    r"""
    Computes the harmonic sum :math:`S_2(N)`.

    .. math::
      S_2(N) = \sum\limits_{j=1}^N \frac 1 {j^2} = -\psi_1(N+1)+\zeta(2)

    with :math:`\psi_1(N)` the trigamma function and :math:`\zeta` the
    Riemann zeta function.

    Parameters
    ----------
      N : complex
        Mellin moment

    Returns
    -------
      S_2 : complex
        Harmonic sum :math:`S_2(N)`

    See Also
    --------
      cern_polygamma : :math:`\psi_0(N)`
    """
    return -cern_polygamma(N + 1.0, 1) + zeta2


@nb.njit("c16(c16)", cache=True)
def harmonic_S3(N):
    r"""
    Computes the harmonic sum :math:`S_3(N)`.

    .. math::
      S_3(N) = \sum\limits_{j=1}^N \frac 1 {j^3} = \frac 1 2 \psi_2(N+1)+\zeta(3)

    with :math:`\psi_2(N)` the 2nd-polygamma function and :math:`\zeta` the
    Riemann zeta function.

    Parameters
    ----------
      N : complex
        Mellin moment

    Returns
    -------
      S_3 : complex
        Harmonic sum :math:`S_3(N)`

    See Also
    --------
      cern_polygamma : :math:`\psi_0(N)`
    """
    return 0.5 * cern_polygamma(N + 1.0, 2) + zeta3


@nb.njit("c16(c16)", cache=True)
def mellin_g3(N):
    r"""
    Computes the Mellin transform of :math:`\text{Li}_2(x)/(1+x)`.

    This function appears in the analytic continuation of the harmonic sum
    :math:`S_{-2,1}(N)` which in turn appears in the NLO anomalous dimension
    (see :ref:`theory/mellin:harmonic sums`).

    Parameters
    ----------
      N : complex
        Mellin moment

    Returns
    -------
      mellin_g3 : complex
        approximate Mellin transform :math:`\mathcal{M}[\text{Li}_2(x)/(1+x)](N)`

    Note
    ----
      We use the name from :cite:`MuselliPhD`, but not his implementation - rather we use the
      Pegasus :cite:`Vogt:2004ns` implementation.
    """
    cs = [1.0000e0, -0.9992e0, 0.9851e0, -0.9005e0, 0.6621e0, -0.3174e0, 0.0699e0]
    g3 = 0
    for j, c in enumerate(cs):
        Nj = N + j
        g3 += c * (zeta2 - harmonic_S1(Nj) / Nj) / Nj
    return g3



# -*- coding: utf-8 -*-
"""
This module contains the definitions of the different basis used, i.e.:

- flavor basis::

    22  -6  -5  -4  -3  -2  -1   21  1   2   3   4   5   6
    gm  tb  bb  cb  sb  ub  db   g   d   u   s   c   b   t

- QCD Evolution basis::

    0   1   2   3   4   5   6   7   8   9  10  11  12  13
    gm  S   g   V   V3  V8 V15 V24 V35  T3 T8  T15 T24 T35

"""

import copy

import numpy as np

flavor_basis_pids = [22] + list(range(-6, -1 + 1)) + [21] + list(range(1, 6 + 1))

evol_basis = [
    "ph",
    "S",
    "g",
    "V",
    "V3",
    "V8",
    "V15",
    "V24",
    "V35",
    "T3",
    "T8",
    "T15",
    "T24",
    "T35",
]

# Tranformation from physical basis to QCD evolution basis
rotate_flavor_to_evolution = np.array(
    [
        [1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
        [0, 1, 1, 1, 1, 1, 1, 0, 1, 1, 1, 1, 1, 1],
        [0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0],
        [0, -1, -1, -1, -1, -1, -1, 0, 1, 1, 1, 1, 1, 1],
        [0, 0, 0, 0, -0, -1, 1, 0, -1, 1, 0, -0, 0, 0],
        [0, 0, 0, 0, 2, -1, -1, 0, 1, 1, -2, -0, 0, 0],
        [0, 0, 0, 3, -1, -1, -1, 0, 1, 1, 1, -3, 0, 0],
        [0, 0, 4, -1, -1, -1, -1, 0, 1, 1, 1, 1, -4, 0],
        [0, 5, -1, -1, -1, -1, -1, 0, 1, 1, 1, 1, 1, -5],
        [0, 0, 0, -0, 0, 1, -1, 0, -1, 1, 0, -0, 0, 0],
        [0, 0, 0, -0, -2, 1, 1, 0, 1, 1, -2, -0, 0, 0],
        [0, 0, 0, -3, 1, 1, 1, 0, 1, 1, 1, -3, 0, 0],
        [0, 0, -4, 1, 1, 1, 1, 0, 1, 1, 1, 1, -4, 0],
        [0, -5, 1, 1, 1, 1, 1, 0, 1, 1, 1, 1, 1, -5],
    ]
)

# inverse transformation
rotate_evolution_to_flavor = np.linalg.inv(rotate_flavor_to_evolution)


def generate_input_from_lhapdf(lhapdf, xs, Q2init):
    """
    Rotate lhapdf-like input object from flavor space to evolution space

    Parameters
    ----------
        lhapdf : object
            lhapdf-like input object (i.e. an object with xfxQ2)
        xs : list
            input x values
        Q2init : float
            input scale

    Returns
    -------
        input_dict : dict
            a mapping evolution element to callable
    """
    # query all pdfs
    flavor_list = []
    empty_pids = []
    for pid in flavor_basis_pids:
        if lhapdf.hasFlavor(pid):
            ls = [lhapdf.xfxQ2(pid, x, Q2init) / x for x in xs]
        else:
            ls = np.zeros(len(xs))
            empty_pids.append(pid)
        flavor_list.append(ls)
    # iterate all pdfs
    input_dict = {}
    for j, evol in enumerate(evol_basis):
        # is it a quark pdf and is it trivial? then skip
        if len(evol) > 1 and evol[0] in ["V", "T"]:
            q = int(np.sqrt(int(evol[1:]) + 1))
            if q in empty_pids and -q in empty_pids:
                continue
        # rotate
        evol_map = rotate_flavor_to_evolution[j]
        input_dict[evol] = evol_map @ flavor_list
    return input_dict


def fill_trivial_dists(old_evols):
    """
    Insert trivial evolutions basis elements.

    Parameters
    ----------
        old_evols : dict
            :class:`eko.evolution_operator.PhysicalOperator`-like dictionary

    Returns
    -------
        evols : dict
            updated dictionary
        trivial_dists : list
            inserted elements
    """
    evols = copy.deepcopy(old_evols)
    # check dependencies beforehand
    if "S" not in evols:
        raise KeyError("No S distribution available")
    if "V" not in evols:
        raise KeyError("No V distribution available")
    trivial_dists = []
    for evol in evol_basis:
        # only rotate in quark distributions
        if evol == "g":
            continue
        # are the target distributions there?
        if evol in ["S", "V"]:
            continue
        # PDF is set?
        if evol in evols:
            continue
        # insert empty photon
        if evol == "ph":
            evols[evol] = np.zeros(len(evols["S"]))
        # insert trivial value
        elif evol[0] == "V":
            evols[evol] = evols["V"]
        elif evol[0] == "T":
            evols[evol] = evols["S"]
        # register for later use
        trivial_dists.append(evol)
    return evols, trivial_dists


def rotate_output(in_evols):
    """
    Rotate lists in evolution basis back to flavor basis.

    Parameters
    ----------
        in_evols : dict
            :class:`eko.evolution_operator.PhysicalOperator`-like dictionary

    Returns
    -------
        out : dict
            rotated dictionary
    """
    # prepare
    evols, trivial_dists = fill_trivial_dists(in_evols)
    evol_list = np.array([evols[evol] for evol in evol_basis])
    final_quark_pid = 6
    final_valence_pid = 6
    for q in range(6, 3, -1):
        if f"T{q*q-1}" in trivial_dists:
            final_quark_pid -= 1
            # assume Vxx too vanish
            final_valence_pid -= 1
            continue
        if f"V{q*q-1}" in trivial_dists:
            final_valence_pid -= 1
    # rotate
    out = {}
    for j, pid in enumerate(flavor_basis_pids):
        # do we actually have to do smth?
        is_non_trivial_ph = pid == 22 and "ph" not in trivial_dists
        is_non_trivial_q = -final_valence_pid <= pid <= final_quark_pid
        if is_non_trivial_ph or is_non_trivial_q or pid == 21:
            flavor_map = rotate_evolution_to_flavor[j]
            out[pid] = flavor_map @ evol_list
    # recover the bared distributions, e.g. tbar = t
    for q in range(-final_quark_pid, -final_valence_pid):
        out[q] = out[-q]
    return out



# -*- coding: utf-8 -*-
r"""
    This module holds the :class:`FlavourTarget` class and the data
    related to the flavour combinations.

    The write-up of the matching conditions is given in
    :doc:`Matching Conditions </Theory/Matching>`.

"""

MINIMAL_NF = 3

##### This can go to a separate file but for now it is ok here
# These are basically the parameters of the FlavourTarget class
# When nothing is given it is assumed nf_min = 3 and original == name
NSV = "NS_v"
NSP = "NS_p"
NSM = "NS_m"
VFNS = {
    "V": (4 * [NSV],),
    "V3": (4 * [NSM],),
    "V8": (4 * [NSM],),
    "V15": (3 * [NSM] + [NSV], "V", 4),
    "V24": (2 * [NSM] + 2 * [NSV], "V", 5),
    "V35": ([NSM] + 3 * [NSV], "V", 6),
    "T3": (4 * [NSP],),
    "T8": (4 * [NSP],),
    "T15": (3 * [NSP], ["S", "g"], 4),
    "T24": (2 * [NSP], ["S", "g"], 5),
    "T35": ([NSP], ["S", "g"], 6),
    "S": ([], ["S", "g"]),
    "g": ([], ["S", "g"]),
}


class FlavourTarget:
    """
    Defines the path across the quark thresholds.

    See :doc:`Matching Conditions </Theory/Matching>`.

    Parameters
    ----------
        name: str
            name of the flavour target (T8, V8, etc)
        path: list(str)
            path to get to the target from the origin at the minimum possible nf
        original: str or list(str)
            original flavour name (or names if the result is a combination)
        nf_min: int
            minimal nf for which this flavour is active
    """

    def __init__(self, name, path, original=None, nf_min=None):
        self.name = name
        self._path = path
        self.nf_0 = MINIMAL_NF
        if nf_min is None:
            nf_min = self.nf_0
        if original is None:
            original = name
        self.nf_min = nf_min
        # In the most general case, the original input can be a combination
        self.base_flavour = original
        # And, actually, gluon and singlet are two special cases
        # TODO this is just a hack because I'm not able to generalize this part
        if name in ["S", "g"]:
            self.force_combination = True
        else:
            self.force_combination = False

    def _path_from_nf(self, nf_target, n_thres):
        """
        Generate the path array from the known nf target
        taking into account nf target == self.nf_0 would be
        the last member of the array

        Parameters
        ----------
            nf_target : int
                number of flavours at target scale
            n_thres : int
                numbers of thresholds to cross

        Returns
        -------
            path : list
                path to get there
        """
        max_nns = len(self._path) + self.nf_0 - 1
        idx_ini = max(max_nns - nf_target, 0)
        idx_fin = idx_ini + n_thres + 1
        return self._path[idx_ini:idx_fin]

    def get_path(self, nf_target, n_thresholds):
        """
        Get the path to a given value of nf
        given a number of thresholds to be crossed

        Parameters
        ----------
            nf_target : int
                nf value of the target flavour
            n_thresholds : int
                number of thresholds which are going to be crossed

        Returns
        -------
            instructions : dict
                a dictonary whose keys are the incoming flavour
                and whose items are the corresponding path
        """
        # TODO can this be made more concise, or is this really the best that can be done?
        # Check from what nf we are coming from
        nf_from = nf_target - n_thresholds
        if nf_from < self.nf_0:
            raise ValueError(
                f"Physical configurations with less than {self.nf_0} were not considered"
            )
        # Now check from which flavour are we coming from
        if nf_from < self.nf_min or self.force_combination:
            flavour_from = self.base_flavour
        else:
            flavour_from = self.name
        # And now check whether this is a single flavour or a combination
        if isinstance(flavour_from, str):
            # Good, trivial
            return_path = [self._path_from_nf(nf_target, n_thresholds)]
            instructions = {flavour_from: return_path}
        else:  # oh, no...
            # Compute the depth of the singlet part of the path
            max_depth = n_thresholds + 1
            if self.force_combination:
                depth = max_depth
            else:
                depth = min(max_depth, self.nf_min - nf_from)
            # Find out the NS part of the path
            shift_th = n_thresholds - depth
            shift_nf = nf_target - depth
            ns_path = self._path_from_nf(shift_nf, shift_th)
            if self.name == "g":
                target_f = "g"
            else:
                target_f = "q"
            instructions = {}
            for flav in flavour_from:
                if flav == "S":
                    from_f = "q"
                elif flav == "g":
                    from_f = "g"
                paths = get_singlet_paths(target_f, from_f, depth)
                # Now insert in the paths the 'trivial' part
                for p in paths:
                    for extra in ns_path:
                        p.insert(0, extra)
                instructions[flav] = paths
        return instructions


def get_singlet_paths(to, fromm, depth):
    """
    Compute all possible path in the singlet sector to reach `to` starting from  `fromm`.

    Parameters
    ----------
        to : 'q' or 'g'
            final point
        fromm : 'q' or 'g'
            starting point
        depth : int
            nesting level; 1 corresponds to the trivial first step

    Returns
    -------
        ls : list(list(str))
            list of all possible paths, where each path is in increasing order, e.g.
            [P1(c <- a), P2(c <- a), ...] and P1(c <- a) = [(c <- b), (b <- a)]
    """
    if depth < 1:
        raise ValueError(f"Invalid arguments: depth >= 1, but got {depth}")
    if to not in ["q", "g"]:
        raise ValueError(f"Invalid arguments: to in [q,g], but got {to}")
    if fromm not in ["q", "g"]:
        raise ValueError(f"Invalid arguments: fromm in [q,g], but got {fromm}")
    # trivial?
    if depth == 1:
        return [[f"S_{to}{fromm}"]]
    # do recursion (if necessary, we could switch back to loops instead)
    qs = get_singlet_paths(to, "q", depth - 1)
    for q in qs:
        q.append(f"S_q{fromm}")
    gs = get_singlet_paths(to, "g", depth - 1)
    for g in gs:
        g.append(f"S_g{fromm}")
    return qs + gs


def get_all_flavour_paths(nf):
    """
    Builds all :class:`FlavourTarget` - see :doc:`Matching Conditions </Theory/Matching>`.

    Parameters
    ----------
        nf : None | int
            if given, paths are filtered by beeing active for at least `nf` flavours

    Returns
    -------
        ls : list(FlavourTarget)
            all instructions
    """
    ls = []
    # build flavour targets
    for flavour, data in VFNS.items():
        # Get the path
        path = data[0]
        base_flavour = None
        min_nf = None
        # Get the base_flavour if it is not itself
        if len(data) > 1:
            base_flavour = data[1]
        # Check at what nf this flavour enters for the first time
        if len(data) > 2:
            min_nf = data[2]
            if nf is not None and min_nf > nf:
                # Skip impossible values
                continue
        flt = FlavourTarget(flavour, path, original=base_flavour, nf_min=min_nf)
        ls.append(flt)
    return ls



# -*- coding: utf-8 -*-
from numbers import Number
import numpy as np


class OpMember:
    """
    A single operator for a specific element in evolution basis.

    This class provide some basic mathematical operations such as products.
    It can also be applied to a pdf vector by the `__call__` method.
    This class will never be exposed to the outside, but will be an internal member
    of the :class:`Operator` and :class:`PhysicalOperator` instances.

    Parameters
    ----------
        value : np.array
            operator matrix
        error : np.array
            operator error matrix
        name : str
            operator name
    """

    def __init__(self, value, error, name):
        self.value = np.array(value)
        self.error = np.array(error)
        self.name = name

    def _split_name(self):
        """Splits the name according to target.input"""
        # we need to do this late, as in raw mode the name to not follow this principle
        name_spl = self.name.split(".")
        if len(name_spl) != 2:
            raise ValueError("The operator name has no valid format: target.input")
        for k in [0, 1]:
            name_spl[k] = name_spl[k].strip()
            if len(name_spl[k]) <= 0:
                raise ValueError("The operator name has no valid format: target.input")
        return name_spl

    @property
    def target(self):
        """Returns target flavour name (given by the first part of the name)"""
        return self._split_name()[0]

    @property
    def input(self):
        """Returns input flavour name (given by the second part of the name)"""
        return self._split_name()[1]

    @property
    def is_physical(self):
        """Lives inside a :class:`PhysicalOperator`? determined by name"""
        for n in ["NS_p", "NS_m", "NS_v", "S_qq", "S_qg", "S_gq", "S_gg"]:
            if self.name.find(n) >= 0:
                return False
        return True

    def apply_pdf(self, pdf_member):
        """
        The operator member can act on a pdf member.

        Parameters
        ----------
            pdf_member : numpy.ndarray
                pdf vector

        Returns
        -------
            result : float
                higher scale pdf
            error : float
                evolution uncertainty to pdf at higher scale
        """
        result = np.dot(self.value, pdf_member)
        error = np.dot(self.error, pdf_member)
        return result, error

    def __str__(self):
        return self.name

    def __mul__(self, operator_member):
        # scalar multiplication
        if isinstance(operator_member, Number):
            one = np.identity(len(self.value))
            rval = operator_member * one
            rerror = 0.0 * one
            new_name = self.name
        # matrix multiplication
        elif isinstance(operator_member, OpMember):
            # check compatibility
            if self.is_physical != operator_member.is_physical:
                raise ValueError("Operators do not live in the same space!")
            if self.is_physical:
                if self.input != operator_member.target:
                    raise ValueError(
                        f"Can not sum {operator_member.name} and {self.name} OpMembers!"
                    )
                new_name = f"{self.target}.{operator_member.input}"
            else:
                new_name = f"{self.name}.{operator_member.name}"
            rval = operator_member.value
            rerror = operator_member.error
        else:
            raise NotImplementedError(
                f"Can't multiply OpMember and {type(operator_member)}"
            )
        lval = self.value
        ler = self.error
        new_val = np.matmul(lval, rval)
        # TODO check error propagation
        new_err = np.abs(np.matmul(lval, rerror)) + np.abs(np.matmul(ler, rval))
        return OpMember(new_val, new_err, new_name)

    def __add__(self, operator_member):
        if isinstance(operator_member, Number):
            # we only allow the integer 0 as alias for the true zero operator
            if operator_member != 0:
                raise ValueError(
                    "The only integer we can sum to is 0 (as alias for the zero operator)"
                )
            rval = operator_member
            rerror = 0.0
            new_name = self.name
        elif isinstance(operator_member, OpMember):
            # check compatibility
            if self.is_physical != operator_member.is_physical:
                raise ValueError("Operators do not live in the same space!")
            if self.is_physical and operator_member.name != self.name:
                raise ValueError(
                    f"Can not sum {operator_member.name} and {self.name} OpMembers!"
                )
            rval = operator_member.value
            rerror = operator_member.error
            new_name = self.name
        else:
            raise NotImplementedError(f"Can't sum OpMember and {type(operator_member)}")
        new_val = self.value + rval
        new_err = self.error + rerror
        return OpMember(new_val, new_err, new_name)

    def __neg__(self):
        return self.__mul__(-1)

    def __eq__(self, operator_member):
        return np.allclose(self.value, operator_member.value)

    def __sub__(self, operator_member):
        return self.__add__(-operator_member)

    def __radd__(self, operator_member):
        return self.__add__(operator_member)

    def __rsub__(self, operator_member):
        return self.__radd__(-operator_member)

    def __rmul__(self, operator_member):
        return self.__mul__(operator_member)

    @staticmethod
    def join(steps, list_of_paths):
        """
        Multiply a list of :class:`OpMember` using the given paths.

        Parameters
        ----------
            steps : list(list(OpMember))
                list of raw operators, with the lowest scale to the right
            list_of_paths : list(list(str))
                list of paths

        Returns
        -------
            final_op : OpMember
                joined operator
        """
        final_op = 0
        for path in list_of_paths:
            cur_op = None
            for step, member in zip(steps, path):
                new_op = step[member]
                if cur_op is None:
                    cur_op = new_op
                else:
                    cur_op = cur_op * new_op
            final_op += cur_op
        return final_op



# -*- coding: utf-8 -*-
"""
    This file contains the output management
"""
import logging
import copy

import numpy as np
import yaml

from . import interpolation
from .operator.member import OpMember
from .operator.physical import PhysicalOperator
from . import basis_rotation as br

logger = logging.getLogger(__name__)


class Output(dict):
    """
    Wrapper for the output to help with application
    to PDFs and dumping to file.
    """

    def apply_pdf(self, lhapdf_like, targetgrid=None, rotate_to_flavor_basis=True):
        """
        Apply all available operators to the input PDFs.

        Parameters
        ----------
            lhapdf_like : object
                object that provides an xfxQ2 callable (as lhapdf, ToyLH does) (and thus is
                in flavor basis)
            targetgrid : list
                if given, interpolates to the pdfs given at targetgrid (instead of xgrid)
            rotate_to_flavor_basis : boolean
                rotate output back to flavor basis

        Returns
        -------
            out_grid : dict
                output PDFs and their associated errors for the computed Q2grid
        """
        # turn lhapdf_like into list
        input_lists = br.generate_input_from_lhapdf(
            lhapdf_like, self["interpolation_xgrid"], self["q2_ref"]
        )

        # build output
        out_grid = {}
        for q2 in self["Q2grid"]:
            pdfs, errs = self.get_op(q2).apply_pdf(input_lists)
            out_grid[q2] = {"pdfs": pdfs, "errors": errs}

        # interpolate to target grid
        if targetgrid is not None:
            b = interpolation.InterpolatorDispatcher.from_dict(self, False)
            rot = b.get_interpolation(targetgrid)
            for q2 in out_grid:
                for pdf_label in out_grid[q2]["pdfs"]:
                    out_grid[q2]["pdfs"][pdf_label] = np.matmul(
                        rot, out_grid[q2]["pdfs"][pdf_label]
                    )
                    out_grid[q2]["errors"][pdf_label] = np.matmul(
                        rot, out_grid[q2]["errors"][pdf_label]
                    )
        # rotate?
        if rotate_to_flavor_basis:
            out_grid = self.rotate_pdfs_to_flavor_basis(out_grid)
        return out_grid

    @staticmethod
    def rotate_pdfs_to_flavor_basis(in_grid):
        """
        Rotate all PDFs from evolution basis to flavor basis

        Parameters
        ----------
            in_grid : dict
                a map q2 to pdfs and their errors in evolution basis

        Returns
        -------
            out_grid : dict
                updated map
        """
        out_grid = {}
        for q2, res in in_grid.items():
            out_grid[q2] = {k: br.rotate_output(v) for k, v in res.items()}
        return out_grid

    def get_raw(self):
        """
        Serialize result as dict/YAML.

        This maps the original numpy matrices to lists.

        Returns
        -------
            out : dict
                dictionary which will be written on output
        """
        # prepare output dict
        out = {
            "Q2grid": {},
        }
        # dump raw elements
        for f in ["interpolation_polynomial_degree", "interpolation_is_log", "q2_ref"]:
            out[f] = self[f]
        # make raw lists
        for k in ["interpolation_xgrid"]:
            out[k] = self[k].tolist()
        # make operators raw
        for q2 in self["Q2grid"]:
            out["Q2grid"][q2] = self.get_op(q2).to_raw()
        return out

    def dump_yaml(self, stream=None):
        """
        Serialize result as YAML.

        Parameters
        ----------
            stream : (Default: None)
                if given, dump is written on it

        Returns
        -------
            dump : any
                result of dump(output, stream), i.e. a string, if no stream is given or
                Null, if self is written sucessfully to stream
        """
        # TODO explicitly silence yaml
        out = self.get_raw()
        return yaml.dump(out, stream)

    def dump_yaml_to_file(self, filename):
        """
        Writes YAML representation to a file.

        Parameters
        ----------
            filename : string
                target file name

        Returns
        -------
            ret : any
                result of dump(output, stream), i.e. Null if written sucessfully
        """
        with open(filename, "w") as f:
            ret = self.dump_yaml(f)
        return ret

    @classmethod
    def load_yaml(cls, stream):
        """
        Load YAML representation from stream

        Parameters
        ----------
            stream : any
                source stream

        Returns
        -------
            obj : output
                loaded object
        """
        obj = yaml.safe_load(stream)
        return cls(obj)

    @classmethod
    def load_yaml_from_file(cls, filename):
        """
        Load YAML representation from file

        Parameters
        ----------
            filename : string
                source file name

        Returns
        -------
            obj : output
                loaded object
        """
        obj = None
        with open(filename) as o:
            obj = Output.load_yaml(o)
        return obj

    def get_op(self, q2):
        """
        Load a :class:`PhysicalOperator` from the raw data.

        Parameters
        ----------
            q2 : float
                target scale

        Returns
        -------
            op : PhysicalOperator
                corresponding Operator
        """
        # check existence
        if q2 not in self["Q2grid"]:
            raise KeyError(f"q2={q2} not in grid")
        # compose
        ops = self["Q2grid"][q2]
        op_members = {}
        for name in ops["operators"]:
            op_members[name] = OpMember(
                ops["operators"][name], ops["operator_errors"][name], name
            )
        return PhysicalOperator(op_members, q2)

    def concat(self, other):
        """
        Concatenate (multiply) two outputs.

        Parameters
        ----------
            other : Output
                other factor to be multiplied from the left, i.e. with *smaller* intial scale

        Returns
        -------
            out : Output
                self * other
        """
        # check type
        if not isinstance(other, Output):
            raise ValueError("can only concatenate two Output instances!")
        # check parameters
        for f in ["interpolation_polynomial_degree", "interpolation_is_log"]:
            if not self[f] == other[f]:
                raise ValueError(
                    f"'{f}' of the two factors does not match: {self[f]} vs {other[f]}"
                )
        if not np.allclose(self["interpolation_xgrid"], other["interpolation_xgrid"]):
            raise ValueError("'interpolation_xgrid' of the two factors does not match")
        # check matching
        mid_scale = self["q2_ref"]
        if not mid_scale in other["Q2grid"]:
            raise ValueError("Operators can not be joined")
        # prepare output
        other_op = other.get_op(mid_scale)
        out = copy.deepcopy(self)
        out["q2_ref"] = other["q2_ref"]
        for q2 in self["Q2grid"]:
            # multiply operators
            me = self.get_op(q2)
            prod = me * other_op
            out["Q2grid"][q2] = prod.to_raw()

        return out



# -*- coding: utf-8 -*-
r"""
Integrals needed for the exact evolutions.

.. math::
    j^{(n,m)}(a_s,a_s^0) = \int\limits_{a_s^0}^{a_s}\!da_s'\,\frac{(a_s')^{1+n}}{-\beta^{(m)}(a_s')}
"""

import numpy as np

import numba as nb

from .. import beta


@nb.njit("f8(f8,f8,u1)", cache=True)
def j00(a1, a0, nf):
    r"""
    LO-LO exact evolution integral.

    .. math::
        j^{(0,0)}(a_s,a_s^0) = \int\limits_{a_s^0}^{a_s} \frac{da_s'}{\beta_0 a_s'}
           = \frac{\ln(a_s/a_s^0)}{\beta_0}

    Parameters
    ----------
        a1 : float
            target coupling value
        a0 : float
            initial coupling value
        nf : int
            number of active flavors

    Returns
    -------
        j00 : float
            integral
    """
    return np.log(a1 / a0) / beta.beta(0, nf)


@nb.njit("f8(f8,f8,u1)", cache=True)
def j11_exact(a1, a0, nf):
    r"""
    NLO-NLO exact evolution integral.

    .. math::
        j^{(1,1)}(a_s,a_s^0) = \int\limits_{a_s^0}^{a_s}\!da_s'\,
                                \frac{a_s'^2}{\beta_0 a_s'^2 + \beta_1 a_s'^3}
            = \frac{1}{\beta_1}\ln\left(\frac{1+b_1 a_s}{1+b_1 a_s^0}\right)

    Parameters
    ----------
        a1 : float
            target coupling value
        a0 : float
            initial coupling value
        nf : int
            number of active flavors

    Returns
    -------
        j11 : float
            integral
    """
    beta_1 = beta.beta(1, nf)
    b1 = beta.b(1, nf)
    return (1.0 / beta_1) * np.log((1.0 + a1 * b1) / (1.0 + a0 * b1))


@nb.njit("f8(f8,f8,u1)", cache=True)
def j11_expanded(a1, a0, nf):
    r"""
    NLO-NLO expanded evolution integral.

    .. math::
        j^{(1,1)}_{exp}(a_s,a_s^0) = \frac 1 {\beta_0}(a_s - a_s^0)

    Parameters
    ----------
        a1 : float
            target coupling value
        a0 : float
            initial coupling value
        nf : int
            number of active flavors

    Returns
    -------
        j11_exp : float
            integral
    """
    return 1.0 / beta.beta(0, nf) * (a1 - a0)


@nb.njit("f8(f8,f8,u1)", cache=True)
def j01_exact(a1, a0, nf):
    r"""
    LO-NLO exact evolution integral.

    .. math::
        j^{(0,1)}(a_s,a_s^0) = \int\limits_{a_s^0}^{a_s}\!da_s'\,
                            \frac{a_s'}{\beta_0 a_s'^2 + \beta_1 a_s'^3}
               = j^{(0,0)}(a_s,a_s^0) - b_1 j^{(1,1)}(a_s,a_s^0)

    Parameters
    ----------
        a1 : float
            target coupling value
        a0 : float
            initial coupling value
        nf : int
            number of active flavors

    Returns
    -------
        j11 : float
            integral
    """
    return j00(a1, a0, nf) - beta.b(1, nf) * j11_exact(a1, a0, nf)


@nb.njit("f8(f8,f8,u1)", cache=True)
def j01_expanded(a1, a0, nf):
    r"""
    LO-NLO expanded evolution integral.

    .. math::
        j^{(0,1)}_{exp}(a_s,a_s^0) = j^{(0,0)}(a_s,a_s^0) - b_1 j^{(1,1)}_{exp}(a_s,a_s^0)

    Parameters
    ----------
        a1 : float
            target coupling value
        a0 : float
            initial coupling value
        nf : int
            number of active flavors

    Returns
    -------
        j01_exp : float
            integral
    """
    return j00(a1, a0, nf) - beta.b(1, nf) * j11_expanded(a1, a0, nf)



# -*- coding: utf-8 -*-
from .member import OpMember


class PhysicalOperator:
    """
    This is exposed to the outside world.

    This operator is computed via the composition method of the
    :class:`Operator` class.

    This operator can act on PDFs through the `__call__` method.


    Parameters
    ----------
        op_members : dict
            list of all members
        q2_final : float
            final scale
    """

    def __init__(self, op_members, q2_final):
        self.op_members = op_members
        self.q2_final = q2_final

    def __mul__(self, other):
        """
        Multiply ``other`` to self.

        Parameters
        ----------
            other : PhysicalOperator
                second factor with a lower initial scale

        Returns
        -------
            p : PhysicalOperator
                self * other
        """
        if not isinstance(other, PhysicalOperator):
            raise ValueError("Can only multiply with another PhysicalOperator")
        # prepare paths
        instructions = {}
        for my_op in self.op_members.values():
            for other_op in other.op_members.values():
                # ops match?
                if my_op.input != other_op.target:
                    continue
                new_key = my_op.target + "." + other_op.input
                path = [my_op.name, other_op.name]
                # new?
                if not new_key in instructions:
                    instructions[new_key] = [path]
                else:  # add element
                    instructions[new_key].append(path)

        # prepare operators
        op_to_compose = [self.op_members] + [other.op_members]
        # iterate operators
        new_ops = {}
        for key, paths in instructions.items():
            new_ops[key] = OpMember.join(op_to_compose, paths)
        return self.__class__(new_ops, self.q2_final)

    def to_raw(self):
        """
        Returns serializable matrix representation of all members and their errors

        Returns
        -------
            ret : dict
                the members are stored under the ``operators`` key and their
                errors under the ``operator_errors`` key. They are labeled as
                ``{outputPDF}.{inputPDF}``.
        """
        # map matrices
        ret = {"operators": {}, "operator_errors": {}}
        for name, op in self.op_members.items():
            ret["operators"][name] = op.value.tolist()
            ret["operator_errors"][name] = op.error.tolist()
        return ret

    def apply_pdf(self, pdf_lists):
        """
        Apply PDFs to the EKOs.

        It assumes as input the PDFs as dictionary in evolution basis with:

        .. code-block:: python

            pdf_lists = {
                'V' : list,
                'g' : list,
                # ...
            }

        Each member has to be evaluated on the corresponding xgrid (which
        is tracked by :class:`~eko.operator_grid.OperatorGrid` and not
        :class:`PhysicalOperator`)

        Parameters
        ----------
            ret : dict
                operator matrices of :class:`PhysicalOperator`
            pdf_lists : dict
                PDFs in evolution basis as list on the corresponding xgrid

        Returns
        -------
            out : dict
                evolved PDFs
            out_errors : dict
                associated errors of the evolved PDFs
        """
        # build output
        outs = {}
        out_errors = {}
        for op in self.op_members.values():
            target, input_pdf = op.target, op.input
            # basis vector available?
            if input_pdf not in pdf_lists:
                # thus can I not complete the calculation for this target
                outs[target] = None
                continue
            # is target new?
            if target not in outs:
                # set output
                outs[target], out_errors[target] = op.apply_pdf(pdf_lists[input_pdf])
            else:
                # is target already blocked?
                if outs[target] is None:
                    continue
                # else add to it
                out, err = op.apply_pdf(pdf_lists[input_pdf])
                outs[target] += out
                out_errors[target] += err
        # remove uncompleted
        outs = {k: outs[k] for k in outs if not outs[k] is None}
        return outs, out_errors



# -*- coding: utf-8 -*-
"""
This module contains the :class:`OperatorGrid`  and the
:class:`OperatorMaster` class.

The first is the driver class of eko as it is the one that collects all the
previously instantiated information and does the actual computation of the Q2s.
"""

import logging
import numbers
import time

import numpy as np

logger = logging.getLogger(__name__)


class OperatorMaster:
    """
    The :class:`OperatorMaster` is instantiated for a given set of parameters
    and informs the generation of operators.

    Parameters
    ----------
        grid: OperatorGrid
            Instance of :class:`OperatorGrid`
        nf: int
            Value of nf for this :class:`OperatorMaster`
    """

    def __init__(self, config, managers, nf):
        self.config = config
        self.managers = managers
        self.nf = nf

    def get_op(self, q2_from, q2_to, generate=False):
        """
        Given a q2_from and a q2_to, returns a raw operator.
        If the `generate` flag is set to True, the operator will also be computed
        in place.

        Parameters
        ----------
            q2_from: float
                Reference value of q^2
            q2_to: float
                Target value of q^2
            generate: bool
                Whether the operator should be computed (default = False)

        Returns
        -------
            op: eko.operators.Operator
                operator to go from q2_from to q2_to
        """
        # import numba late - that is only here
        start = time.perf_counter()
        logger.info("Evolution: compiling Numba")
        from . import Operator  # pylint: disable=import-outside-toplevel

        logger.info(
            "Evolution: compiling Numba - took %.1f s", time.perf_counter() - start
        )
        # now do the computation
        op = Operator(self.config, self.managers, self.nf, q2_from, q2_to)
        if generate:
            op.compute()
        return op


class OperatorGrid:
    """
    The operator grid is the driver class of the evolution.

    It receives as input a threshold holder and a generator of a_s.
    From that point onwards it can compute any operator at any q2.

    Parameters
    ----------
        q2_grid: array
            Grid in Q2 on where to to compute the operators
        order: int
            order in perturbation theory
        thresholds_config: eko.thresholds.ThresholdsConfig
            Instance of :class:`~eko.thresholds.Threshold` containing information about the
            thresholds
        strong_coupling: eko.strong_coupling.StrongCoupling
            Instance of :class:`~eko.strong_coupling.StrongCoupling` able to generate a_s for
            any q
        kernel_dispatcher: eko.kernel_generation.KernelDispatcher
            Instance of the :class:`~eko.kernel_generation.KernelDispatcher` with the
            information about the kernels
    """

    def __init__(
        self,
        config,
        q2_grid,
        thresholds_config,
        strong_coupling,
        interpol_dispatcher,
    ):
        # check
        order = int(config["order"])
        method = config["method"]
        if not method in [
            "iterate-exact",
            "iterate-expanded",
            "truncated",
            "ordered-truncated",
            "decompose-exact",
            "decompose-expanded",
            "perturbative-exact",
            "perturbative-expanded",
        ]:
            raise ValueError(f"Unknown evolution mode {method}")
        if order == 0 and method != "iterate-exact":
            logger.warning("Evolution: In LO we use the exact solution always!")
        self.config = config
        self.q2_grid = q2_grid
        self.managers = dict(
            thresholds_config=thresholds_config,
            strong_coupling=strong_coupling,
            interpol_dispatcher=interpol_dispatcher,
        )
        self._op_masters = {}
        self._op_grid = {}
        self._threshold_operators = {}

    @classmethod
    def from_dict(
        cls,
        setup,
        thresholds_config,
        strong_coupling,
        interpol_dispatcher,
    ):
        """
        Create the object from the theory dictionary.

        Read keys:

            - Q2grid : required, target grid
            - debug_skip_singlet : optional, avoid singlet integrations?
            - debug_skip_non_singlet : optional, avoid non-singlet integrations?

        Parameters
        ----------
            setup : dict
                theory dictionary
            thresholds_config : eko.thresholds.ThresholdsConfig
                An instance of the ThresholdsConfig class
            strong_coupling : eko.strong_coupling.StrongCoupling
                An instance of the StrongCoupling class
            interpol_dispatcher : eko.interpolation.InterpolatorDispatcher
                An instance of the InterpolatorDispatcher class

        Returns
        -------
            obj : cls
                created object
        """
        config = {}
        config["order"] = int(setup["PTO"])
        method = setup.get("ModEv", "iterate-exact")
        mod_ev2method = {
            "EXA": "iterate-exact",
            "EXP": "iterate-expanded",
            "TRN": "truncated",
        }
        method = mod_ev2method.get(method, method)
        config["method"] = method
        config["fact_to_ren"] = (setup["XIF"] / setup["XIR"]) ** 2
        config["ev_op_max_order"] = setup.get("ev_op_max_order", 10)
        config["ev_op_iterations"] = setup.get("ev_op_iterations", 10)
        config["debug_skip_singlet"] = setup.get("debug_skip_singlet", False)
        config["debug_skip_non_singlet"] = setup.get("debug_skip_non_singlet", False)
        q2_grid = np.array(setup["Q2grid"], np.float_)
        return cls(
            config, q2_grid, thresholds_config, strong_coupling, interpol_dispatcher
        )

    def _generate_masters(self):
        """
        Creates all :class:`OperatorMaster` instances.
        """
        for nf in self.managers["thresholds_config"].nf_range():
            # Set up the OperatorMaster for each nf
            self._op_masters[nf] = OperatorMaster(
                self.config,
                {k: v for k, v in self.managers.items() if k != "thresholds_config"},
                nf,
            )

    def _generate_thresholds_op(self, to_q2):
        """
        Generate the threshold operators

        This method is called everytime the OperatorGrid is asked for a grid on Q^2
        with a list of the relevant areas.
        If new threshold operators need to be computed, they will be
        cached in an internal dictionary.

        The internal dictionary is self._threshold_operators and its structure is:
        (q2_from, q2_to) -> eko.operators.Operator

        Parameters
        ----------
            to_q2: float
                value of q2 for which the OperatorGrid will need to pass thresholds
        """
        # The lists of areas as produced by the thresholds
        area_list = self.managers["thresholds_config"].get_path_from_q2_ref(to_q2)
        # The base area is always that of the reference q
        q2_from = self.managers["thresholds_config"].q2_ref
        nf = self.managers["thresholds_config"].nf_ref
        for area in area_list:
            q2_to = area.q2_ref
            if q2_to == q2_from:
                continue
            new_op = (q2_from, q2_to)
            logger.info(
                "Evolution: Compute threshold operator from %e to %e", q2_from, q2_to
            )
            if new_op not in self._threshold_operators:
                # Compute the operator in place and store it
                op_th = self._op_masters[nf].get_op(q2_from, q2_to, generate=True)
                self._threshold_operators[new_op] = op_th
            nf = area.nf
            q2_from = q2_to

    def _get_jumps(self, qsq):
        """
        Given a value of q^2, generates the list of operators that need to be
        composed in order to get there from q0^2

        Parameters
        ----------
            qsq: float
                Target value of q^2

        Returns
        -------
            op_list: list
                List of threshold operators
        """
        # Get the list of areas to be crossed
        full_area_path = self.managers["thresholds_config"].get_path_from_q2_ref(qsq)
        # The last one is where q resides so it is not needed
        area_path = full_area_path[:-1]
        op_list = []
        # Now loop over the areas to collect the necessary threshold operators
        for area in area_path:
            q2_from = area.q2_ref
            q2_to = area.q2_towards(qsq)
            op_list.append(self._threshold_operators[(q2_from, q2_to)])
        return op_list

    def set_q2_limits(self, q2min, q2max):
        """
        Sets up the limits of the grid in q^2 to be computed by the OperatorGrid

        This function is a wrapper to compute the necessary operators to go between areas

        Parameters
        ----------
            q2min: float
                Minimum value of q^2 that will be computed
            q2max: float
                Maximum value of q^2 that will be computed
        """
        # Sanity checks
        if q2min <= 0.0 or q2max <= 0.0:
            raise ValueError(
                f"Values of q^2 below 0.0 are not accepted, received [{q2min},{q2max}]"
            )
        if q2min > q2max:
            raise ValueError(
                f"Minimum q^2 is above maximum q^2 (error: {q2max} < {q2min})"
            )
        # Ensure we have all the necessary operators to go from q2ref to q2min and qmax
        self._generate_thresholds_op(q2min)
        self._generate_thresholds_op(q2max)

    def _compute_raw_grid(self, q2grid):
        """
        Receives a grid in q^2 and computes each opeator inside its
        area with reference value the q_ref of its area

        Parameters
        ----------
            q2grid: list
                List of q^2
        """
        area_list = self.managers["thresholds_config"].get_areas(q2grid)
        for area, q2 in zip(area_list, q2grid):
            q2_from = area.q2_ref
            nf = area.nf
            logger.info("Evolution: compute operators %e -> %e", q2_from, q2)
            self._op_grid[q2] = self._op_masters[nf].get_op(q2_from, q2)
        # Now perform the computation
        for op in self._op_grid.values():
            op.compute()

    def compute_q2grid(self, q2grid=None):
        """
        Receives a grid in q^2 and computes all operations necessary
        to return any operator at any given q for the evolution between q2ref and q2grid

        Parameters
        ----------
            q2grid: list
                List of q^2

        Returns
        -------
            grid_return: list
                List of PhysicalOperator for each value of q^2
        """
        # use input?
        if q2grid is None:
            q2grid = self.q2_grid
        # normalize input
        if isinstance(q2grid, numbers.Number):
            q2grid = [q2grid]
        # build masters
        if len(self._op_masters) <= 0:
            self._generate_masters()
        # Check max and min of the grid and reset the limits if necessary
        q2max = np.max(q2grid)
        q2min = np.min(q2grid)
        self.set_q2_limits(q2min, q2max)
        # Now compute all raw operators
        self._compute_raw_grid(q2grid)
        # And now return the grid
        grid_return = []
        for q2 in q2grid:
            grid_return.append(self.get_op_at_q2(q2))
        return grid_return

    def get_op_at_q2(self, qsq):
        """
        Given a value of q^2, returns the PhysicalOperator to get
        to q^2 from q0^2

        Parameters
        ----------
            qsq: float
                Target value of q^2

        Returns
        -------
            final_op: eko.evolution_operator.PhysicalOperator
                Op(q^2 <- q_0^2)
        """
        # Check the path to q0 for this operator
        if qsq in self._op_grid:
            operator = self._op_grid[qsq]
        else:
            logger.warning("Evolution: Q2=%e not found in the grid, computing...", qsq)
            self.compute_q2grid(qsq)
            operator = self._op_grid[qsq]
        # Prepare the path for the composition of the operator
        operators_to_q2 = self._get_jumps(operator.q2_from)
        number_of_thresholds = len(operators_to_q2)
        instruction_set = self.managers["thresholds_config"].get_composition_path(
            operator.nf, number_of_thresholds
        )
        # Compose and return
        final_op = operator.compose(operators_to_q2, instruction_set, qsq)
        return final_op



# -*- coding: utf-8 -*-
r"""
This module contains the QCD beta function coefficients.

See :doc:`pQCD ingredients </Theory/pQCD>`.
"""

import numba as nb

from eko import constants


@nb.njit("f8(u1)", cache=True)
def beta_0(nf: int):
    """
    Computes the first coefficient of the QCD beta function.

    Implements Eq. (3.1) of :cite:`Herzog:2017ohr`.

    Parameters
    ----------
        nf : int
            number of active flavours

    Returns
    -------
        beta_0 : float
            first coefficient of the QCD beta function :math:`\\beta_0^{n_f}`
    """
    beta_0 = 11.0 / 3.0 * constants.CA - 4.0 / 3.0 * constants.TR * nf
    return beta_0


@nb.njit("f8(u1)", cache=True)
def beta_1(nf: int):
    """
    Computes the second coefficient of the QCD beta function.

    Implements Eq. (3.2) of :cite:`Herzog:2017ohr`.

    Parameters
    ----------
        nf : int
            number of active flavours

    Returns
    -------
        beta_1 : float
            second coefficient of the QCD beta function :math:`\\beta_1^{n_f}`
    """
    TF = constants.TR * nf
    b_ca2 = 34.0 / 3.0 * constants.CA * constants.CA
    b_ca = -20.0 / 3.0 * constants.CA * TF
    b_cf = -4.0 * constants.CF * TF
    beta_1 = b_ca2 + b_ca + b_cf
    return beta_1


@nb.njit("f8(u1)", cache=True)
def beta_2(nf: int):
    """
    Computes the third coefficient of the QCD beta function

    Implements Eq. (3.3) of :cite:`Herzog:2017ohr`.

    Parameters
    ----------
        nf : int
            number of active flavours

    Returns
    -------
        beta_2 : float
            third coefficient of the QCD beta function :math:`\\beta_2^{n_f}`
    """
    TF = constants.TR * nf
    beta_2 = (
        2857.0 / 54.0 * constants.CA * constants.CA * constants.CA
        - 1415.0 / 27.0 * constants.CA * constants.CA * TF
        - 205.0 / 9.0 * constants.CF * constants.CA * TF
        + 2.0 * constants.CF * constants.CF * TF
        + 44.0 / 9.0 * constants.CF * TF * TF
        + 158.0 / 27.0 * constants.CA * TF * TF
    )
    return beta_2


@nb.njit("f8(u1,u1)", cache=True)
def beta(k, nf):
    """
    Compute value of a beta coefficients

    Parameters
    ----------
        k : int
            perturbative order
        nf : int
            number of active flavours

    Returns
    -------
        beta : float
            beta_k(nf)
    """
    beta_ = 0
    if k == 0:
        beta_ = beta_0(nf)
    elif k == 1:
        beta_ = beta_1(nf)
    elif k == 2:
        beta_ = beta_2(nf)
    else:
        raise ValueError("Beta coefficients beyond NNLO are not implemented!")
    return beta_


@nb.njit("f8(u1,u1)", cache=True)
def b(k, nf):
    """
    Compute b coefficient.

    Parameters
    ----------
        k : int
            perturbative order
        nf : int
            number of active flavours

    Returns
    -------
        b : float
            b_k(nf)
    """
    return beta(k, nf) / beta(0, nf)



# -*- coding: utf-8 -*-
r"""
    This module holds the classes that define the FNS: Threshold,
    Area
"""
import logging
import numbers

import numpy as np

from eko.flavours import get_all_flavour_paths

logger = logging.getLogger(__name__)


class Area:
    """
    Sets up a single threhold area with a fixed configuration.

    Parameters
    ----------
        q2_min : float
            lower bound of the area
        q2_max : float
            upper bound of the area
        q2_0 : float
            reference point of the area (can be anywhere in the area)
        nf : float
            number of flavours in the area
    """

    def __init__(self, q2_min, q2_max, q2_0, nf):
        self.q2_min = q2_min
        self.q2_max = q2_max
        self.nf = nf
        self.has_q2_0 = False
        # Now check which is the q2_ref for this area
        if q2_0 > q2_max:
            self.q2_ref = q2_max
        elif q2_0 < q2_min:
            self.q2_ref = q2_min
        else:
            self.has_q2_0 = True
            self.q2_ref = q2_0

    def q2_towards(self, q2):
        """
        Return q2_min or q2_max depending on whether
        we are going towards the max or the min or q2
        if we are alreay in the correct area

        Parameters
        ----------
            q2 : float
                reference point

        Returns
        -------
            q2_next : float
                the closest point to q2 that is within the area
        """
        if q2 > self.q2_max:
            return self.q2_max
        elif q2 < self.q2_min:
            return self.q2_min
        else:
            return q2

    def __gt__(self, target_area):
        """Compares q2 of areas"""
        return self.q2_min >= target_area.q2_max

    def __lt__(self, target_area):
        """Compares q2 of areas"""
        return self.q2_max <= target_area.q2_min

    def __call__(self, q2):
        """
        Checks whether q2 is contained in the area

        Parameters
        ----------
            q2 : float
                testing point

        Returns
        -------
            contained : bool
                is point contained?
        """
        return self.q2_min <= q2 <= self.q2_max


class ThresholdsConfig:
    """
    The threshold class holds information about the thresholds any
    Q2 has to pass in order to get there from a given q2_ref and scheme.

    Parameters
    ----------
        q2_ref : float
            Reference q^2
        scheme: str
            Choice of scheme
        threshold_list: list
            List of q^2 thresholds should the scheme accept it
        nf: int
            Number of flavors for the FFNS
    """

    def __init__(self, q2_ref, scheme, *, threshold_list=None, nf=None):
        # Initial values
        self.q2_ref = q2_ref
        self.scheme = scheme
        self._threshold_list = []
        self._areas = []
        self._area_walls = []
        self._area_ref = 0
        self.max_nf = None
        self.min_nf = None
        self._operator_paths = []
        protection = False

        if scheme == "FFNS":
            if nf is None:
                raise ValueError("No value for nf in the FFNS was received")
            if threshold_list is not None:
                raise ValueError("The FFNS does not accept any thresholds")
            self._areas = [Area(0, np.inf, self.q2_ref, nf)]
            self.max_nf = nf
            self.min_nf = nf
            protection = True
            logger.info("Thresholds: Fixed flavor number scheme with %d flavors", nf)
        elif scheme in ["ZM-VFNS", "FONLL-A", "FONLL-A'"]:
            if nf is not None:
                logger.warning(
                    "The ZM-VFNS configures its own value for nf, ignoring input nf=%d",
                    nf,
                )
            if threshold_list is None:
                raise ValueError(
                    "The ZM-VFN scheme was selected but no thresholds were given"
                )
            self._setup_vfns(threshold_list)
            logger.info("Thresholds: Variable flavor number scheme (%s)", scheme)
        else:
            raise NotImplementedError(f"The scheme {scheme} is not implemented")

        # build flavour targets
        nf_protected = None
        if protection:
            nf_protected = nf
        self._operator_paths = get_all_flavour_paths(nf_protected)

    @classmethod
    def from_dict(cls, setup):
        FNS = setup["FNS"]
        q2_ref = pow(setup["Q0"], 2)
        if FNS != "FFNS":  # setup ZM-VFNS
            mc = setup["mc"]
            mb = setup["mb"]
            mt = setup["mt"]
            threshold_list = pow(np.array([mc, mb, mt]), 2)
            nf = None
        else:  # here FFNS
            nf = setup["NfFF"]
            threshold_list = None
        return cls(q2_ref, FNS, threshold_list=threshold_list, nf=nf)

    @property
    def nf_ref(self):
        """ Number of flavours in the reference area """
        return self._areas[self._area_ref].nf

    def nf_range(self):
        """
        Iterate number of flavours, including min_nf *and* max_nf

        Yields
        ------
            nf : int
                number of flavour
        """
        return range(self.min_nf, self.max_nf + 1)

    def _setup_vfns(self, threshold_list):
        """
        Receives a list of thresholds and sets up the vfns scheme

        Parameters
        ----------
            threshold_list: list
                List of q^2 thresholds
        """
        nf = 3
        # Force sorting
        self._area_walls = sorted(threshold_list)
        # Generate areas
        self._areas = []
        q2_min = 0
        q2_ref = self.q2_ref
        self.min_nf = nf
        for i, q2_max in enumerate(self._area_walls + [np.inf]):
            nf = self.min_nf + i
            new_area = Area(q2_min, q2_max, q2_ref, nf)
            if new_area.has_q2_0:
                self._area_ref = i
            self._areas.append(new_area)
            q2_min = q2_max
        self.max_nf = nf

    def get_path_from_q2_ref(self, q2):
        """
        Get the Area path from q2_ref to q2.

        Parameters
        ----------
            q2: float
                Target value of q2

        Returns
        -------
            area_path: list
                List of Areas to go through in order to get from q2_ref
                to q2. The first one is the one containg q2_ref while the
                last one contains q2
        """
        current_area = self.get_areas_idx(q2)[0]
        if current_area < self._area_ref:
            rc = -1
        else:
            rc = 1
        area_path = [
            self._areas[i] for i in range(self._area_ref, current_area + rc, rc)
        ]
        return area_path

    def get_composition_path(self, nf, n_thres):
        """
        Iterates all flavour targets.

        Parameters
        ----------
            nf: int
                nf value of the target flavour
            n_thres: int
                number of thresholds which are going to be crossed

        Yields
        ------
            name : string
                flavour name
            path : list
                flavour path
        """
        for flavour in self._operator_paths:
            yield flavour.name, flavour.get_path(nf, n_thres)

    def get_areas_idx(self, q2arr):
        """
        Returns the index of the area in which each value of q2arr falls.

        Parameters
        ----------
            q2arr: np.array
                array of values of q2

        Returns
        -------
            areas_idx: list
                list with the indices of the corresponding areas for q2arr
        """
        # Ensure q2arr is an array
        if isinstance(q2arr, numbers.Number):
            q2arr = np.array([q2arr])
        # Check in which area is every q2
        areas_idx = np.digitize(q2arr, self._area_walls)
        return areas_idx

    def get_areas(self, q2arr):
        """
        Returns the Areas in which each value of q2arr falls

        Parameters
        ----------
            q2arr: np.array
                array of values of q2

        Returns
        -------
            areas: list
                list with the areas for q2arr
        """
        idx = self.get_areas_idx(q2arr)
        area_list = np.array(self._areas)[idx]
        return list(area_list)



# -*- coding: utf-8 -*-
"""
Colletion of singlet EKOs.
"""

import numpy as np

import numba as nb

from .. import beta
from .. import anomalous_dimensions as ad

from . import evolution_integrals as ei
from . import utils


@nb.njit("c16[:,:](c16[:,:,:],f8,f8,u1)", cache=True)
def lo_exact(gamma_singlet, a1, a0, nf):
    """
    Singlet leading order exact EKO

    Parameters
    ----------
        gamma_singlet : numpy.ndarray
            singlet anomalous dimensions matrices
        a1 : float
            target coupling value
        a0 : float
            initial coupling value
        nf : int
            number of active flavors

    Returns
    -------
        e_s^0 : numpy.ndarray
            singlet leading order exact EKO
    """
    return ad.exp_singlet(gamma_singlet[0] * ei.j00(a1, a0, nf))[0]


@nb.njit("c16[:,:](c16[:,:,:],f8,f8)", cache=True)
def nlo_decompose(gamma_singlet, j01, j11):
    """
    Singlet next-to-leading order decompose EKO

    Parameters
    ----------
        gamma_singlet : numpy.ndarray
            singlet anomalous dimensions matrices
        a1 : float
            target coupling value
        a0 : float
            initial coupling value
        nf : int
            number of active flavors
        j01 : float
            LO-NLO evolution integral
        j11 : float
            NLO-NLO evolution integral

    Returns
    -------
        e_s^1 : numpy.ndarray
            singlet next-to-leading order decompose EKO
    """
    return ad.exp_singlet(gamma_singlet[0] * j01 + gamma_singlet[1] * j11)[0]


@nb.njit("c16[:,:](c16[:,:,:],f8,f8,u1)", cache=True)
def nlo_decompose_exact(gamma_singlet, a1, a0, nf):
    """
    Singlet next-to-leading order decompose-exact EKO

    Parameters
    ----------
        gamma_singlet : numpy.ndarray
            singlet anomalous dimensions matrices
        a1 : float
            target coupling value
        a0 : float
            initial coupling value
        nf : int
            number of active flavors

    Returns
    -------
        e_s^1 : numpy.ndarray
            singlet next-to-leading order decompose-exact EKO
    """
    return nlo_decompose(
        gamma_singlet, ei.j01_exact(a1, a0, nf), ei.j11_exact(a1, a0, nf)
    )


@nb.njit("c16[:,:](c16[:,:,:],f8,f8,u1)", cache=True)
def nlo_decompose_expanded(gamma_singlet, a1, a0, nf):
    """
    Singlet next-to-leading order decompose-expanded EKO

    Parameters
    ----------
        gamma_singlet : numpy.ndarray
            singlet anomalous dimensions matrices
        a1 : float
            target coupling value
        a0 : float
            initial coupling value
        nf : int
            number of active flavors

    Returns
    -------
        e_s^1 : numpy.ndarray
            singlet next-to-leading order decompose-expanded EKO
    """
    return nlo_decompose(
        gamma_singlet, ei.j01_expanded(a1, a0, nf), ei.j11_expanded(a1, a0, nf)
    )


@nb.njit("c16[:,:](c16[:,:,:],f8,f8,u1,u4)", cache=True)
def nlo_iterate(gamma_singlet, a1, a0, nf, ev_op_iterations):
    """
    Singlet next-to-leading order iterated (exact) EKO

    Parameters
    ----------
        gamma_singlet : numpy.ndarray
            singlet anomalous dimensions matrices
        a1 : float
            target coupling value
        a0 : float
            initial coupling value
        nf : int
            number of active flavors
        ev_op_iterations : int
            number of evolution steps

    Returns
    -------
        e_s^1 : numpy.ndarray
            singlet next-to-leading order iterated (exact) EKO
    """
    a_steps = utils.geomspace(a0, a1, ev_op_iterations)
    beta0 = beta.beta(0, nf)
    beta1 = beta.beta(1, nf)
    e = np.identity(2, np.complex_)
    al = a_steps[0]
    for ah in a_steps[1:]:
        a_half = (ah + al) / 2.0
        delta_a = ah - al
        ln = (
            (gamma_singlet[0] * a_half + gamma_singlet[1] * a_half ** 2)
            / (beta0 * a_half ** 2 + beta1 * a_half ** 3)
            * delta_a
        )
        ek = np.ascontiguousarray(ad.exp_singlet(ln)[0])
        e = ek @ e
        al = ah
    return e


@nb.njit("c16[:,:,:](c16[:,:,:],u1,u1,b1)", cache=True)
def r_vec(gamma_singlet, nf, ev_op_max_order, is_exact):
    r"""
    Compute singlet R vector for perturbative mode.

    .. math::
        \frac{d}{da_s} \dSV{1}{a_s} &= \frac{\mathbf R (a_s)}{a_s} \cdot \dSV{1}{a_s}\\
        \mathbf R (a_s) &= \sum\limits_{k=0} a_s^k \mathbf R_{k}

    Parameters
    ----------
        gamma_singlet : numpy.ndarray
            singlet anomalous dimensions matrices
        nf : int
            number of active flavors
        ev_op_iterations : int
            number of evolution steps
        is_exact : boolean
            fill up r-vector?

    Returns
    -------
        r : np.ndarray
            R vector
    """
    r = np.zeros((ev_op_max_order + 1, 2, 2), np.complex_)  # k = 0 .. max_order
    beta0 = beta.beta(0, nf)
    b1 = beta.b(1, nf)
    # fill explicit elements
    r[0] = gamma_singlet[0] / beta0
    r[1] = gamma_singlet[1] / beta0 - b1 * r[0]
    # fill rest
    if is_exact:
        for kk in range(2, ev_op_max_order + 1):
            r[kk] = -b1 * r[kk - 1]
    return r


@nb.njit("c16[:,:,:](c16[:,:,:],u1,u1)", cache=True)
def nlo_r_exact(gamma_singlet, nf, ev_op_max_order):
    """
    Compute singlet R vector for perturbative-exact mode.

    Parameters
    ----------
        gamma_singlet : list(numpy.ndarray)
            singlet anomalous dimensions matrices
        nf : int
            number of active flavors
        ev_op_iterations : int
            number of evolution steps

    Returns
    -------
        r : np.ndarray
            R vector

    See Also
    --------
        r_vec : compute R vector
    """
    return r_vec(gamma_singlet, nf, ev_op_max_order, True)


@nb.njit("c16[:,:,:](c16[:,:,:],u1,u1)", cache=True)
def nlo_r_expanded(gamma_singlet, nf, ev_op_max_order):
    """
    Compute singlet R vector for perturbative-expanded mode.

    Parameters
    ----------
        gamma_singlet : list(numpy.ndarray)
            singlet anomalous dimensions matrices
        nf : int
            number of active flavors
        ev_op_iterations : int
            number of evolution steps

    Returns
    -------
        r : np.ndarray
            R vector

    See Also
    --------
        r_vec : compute R vector
    """
    return r_vec(gamma_singlet, nf, ev_op_max_order, False)


@nb.njit("c16[:,:,:](c16[:,:,:],u1)", cache=True)
def u_vec(r, ev_op_max_order):
    r"""
    Compute the elements of the singlet U vector.

    .. math::
        \ESk{n}{a_s}{a_s^0} &= \mathbf U (a_s) \ESk{0}{a_s}{a_s^0} {\mathbf U}^{-1} (a_s^0)\\
        \mathbf U (a_s) &= \mathbf I + \sum\limits_{k=1} a_s^k \mathbf U_k

    Parameters
    ----------
        r : numpy.ndarray
            singlet R vector
        ev_op_max_order : int
            perturbative expansion order of U

    Returns
    -------
        u : np.ndarray
            U vector
    """
    u = np.zeros((ev_op_max_order + 1, 2, 2), np.complex_)  # k = 0 .. max_order
    # init
    u[0] = np.identity(2, np.complex_)
    _, r_p, r_m, e_p, e_m = ad.exp_singlet(r[0])
    e_p = np.ascontiguousarray(e_p)
    e_m = np.ascontiguousarray(e_m)
    for kk in range(1, ev_op_max_order + 1):
        # compute R'
        rp = np.zeros((2, 2), np.complex_)
        for jj in range(kk):
            rp += np.ascontiguousarray(r[kk - jj]) @ u[jj]
        # now compose U
        u[kk] = (
            (e_m @ rp @ e_m + e_p @ rp @ e_p) / kk
            + ((e_p @ rp @ e_m) / (r_m - r_p + kk))
            + ((e_m @ rp @ e_p) / (r_p - r_m + kk))
        )
    return u


@nb.njit("c16[:,:](c16[:,:,:],f8)", cache=True)
def sum_u(uvec, a):
    r"""
    Sums up the actual U operator.


    .. math::
        \mathbf U (a_s) = \mathbf I + \sum\limits_{k=1} a_s^k \mathbf U_k

    Parameters
    ----------
        uvec : numpy.ndarray
            U vector (elements)
        a : float
            strong coupling

    Returns
    -------
        u : numpy.ndarray
            sum
    """
    p = 1.0
    res = np.zeros((2, 2), np.complex_)
    for uk in uvec:
        res += p * uk
        p *= a
    # alternative implementation:
    # al_vec = al**(np.arange(len(uk)))
    # ul = np.sum(al_vec * uk.T,-1).T
    return res


@nb.njit("c16[:,:](c16[:,:,:],f8,f8,u1,u4,u1,c16[:,:,:])", cache=True)
def nlo_perturbative(gamma_singlet, a1, a0, nf, ev_op_iterations, ev_op_max_order, r):
    """
    Singlet next-to-leading order pertubative EKO

    Parameters
    ----------
        gamma_singlet : numpy.ndarray
            singlet anomalous dimensions matrices
        a1 : float
            target coupling value
        a0 : float
            initial coupling value
        nf : int
            number of active flavors
        ev_op_iterations : int
            number of evolution steps
        ev_op_max_order : int
            perturbative expansion order of U
        r : numpy.ndarray
            R vector

    Returns
    -------
        e_s^1 : numpy.ndarray
            singlet next-to-leading order perturbative EKO
    """
    uk = u_vec(r, ev_op_max_order)
    e = np.identity(2, np.complex_)
    # iterate elements
    a_steps = utils.geomspace(a0, a1, ev_op_iterations)
    al = a_steps[0]
    for ah in a_steps[1:]:
        e0 = lo_exact(gamma_singlet, ah, al, nf)
        uh = sum_u(uk, ah)
        ul = sum_u(uk, al)
        # join elements
        ek = np.ascontiguousarray(uh) @ np.ascontiguousarray(e0) @ np.linalg.inv(ul)
        # import pdb; pdb.set_trace()
        e = ek @ e
        al = ah
    return e


@nb.njit("c16[:,:](c16[:,:,:],f8,f8,u1,u4,u1)", cache=True)
def nlo_perturbative_exact(
    gamma_singlet, a1, a0, nf, ev_op_iterations, ev_op_max_order
):
    """
    Singlet next-to-leading order pertubative-exact EKO

    Parameters
    ----------
        gamma_singlet : numpy.ndarray
            singlet anomalous dimensions matrices
        a1 : float
            target coupling value
        a0 : float
            initial coupling value
        nf : int
            number of active flavors
        ev_op_iterations : int
            number of evolution steps
        ev_op_max_order : int
            perturbative expansion order of U

    Returns
    -------
        e_s^1 : numpy.ndarray
            singlet next-to-leading order perturbative-exact EKO

    See Also
    --------
        nlo_perturbative : called function
    """
    r = nlo_r_exact(gamma_singlet, nf, ev_op_max_order)
    return nlo_perturbative(
        gamma_singlet, a1, a0, nf, ev_op_iterations, ev_op_max_order, r
    )


@nb.njit("c16[:,:](c16[:,:,:],f8,f8,u1,u4,u1)", cache=True)
def nlo_perturbative_expanded(
    gamma_singlet, a1, a0, nf, ev_op_iterations, ev_op_max_order
):
    """
    Singlet next-to-leading order pertubative-expanded EKO

    Parameters
    ----------
        gamma_singlet : numpy.ndarray
            singlet anomalous dimensions matrices
        a1 : float
            target coupling value
        a0 : float
            initial coupling value
        nf : int
            number of active flavors
        ev_op_iterations : int
            number of evolution steps
        ev_op_max_order : int
            perturbative expansion order of U

    Returns
    -------
        e_s^1 : numpy.ndarray
            singlet next-to-leading order perturbative-expanded EKO

    See Also
    --------
        nlo_perturbative : called function
    """
    r = nlo_r_expanded(gamma_singlet, nf, ev_op_max_order)
    return nlo_perturbative(
        gamma_singlet, a1, a0, nf, ev_op_iterations, ev_op_max_order, r
    )


@nb.njit("c16[:,:](c16[:,:,:],f8,f8,u1,u4)", cache=True)
def nlo_truncated(gamma_singlet, a1, a0, nf, ev_op_iterations):
    """
    Singlet next-to-leading order truncated EKO

    Parameters
    ----------
        gamma_singlet : numpy.ndarray
            singlet anomalous dimensions matrices
        a1 : float
            target coupling value
        a0 : float
            initial coupling value
        nf : int
            number of active flavors
        ev_op_iterations : int
            number of evolution steps

    Returns
    -------
        e_s^1 : numpy.ndarray
            singlet next-to-leading order truncated EKO
    """
    r = nlo_r_expanded(gamma_singlet, nf, 1)
    u = u_vec(r, 1)
    u1 = np.ascontiguousarray(u[1])
    e = np.identity(2, np.complex_)
    # iterate elements
    a_steps = utils.geomspace(a0, a1, ev_op_iterations)
    al = a_steps[0]
    for ah in a_steps[1:]:
        e0 = np.ascontiguousarray(lo_exact(gamma_singlet, ah, al, nf))
        ek = e0 + ah * u1 @ e0 - al * e0 @ u1
        e = ek @ e
        al = ah
    return e


@nb.njit("c16[:,:](u1,string,c16[:,:,:],f8,f8,u1,u4,u1)", cache=True)
def dispatcher(  # pylint: disable=too-many-return-statements
    order, method, gamma_singlet, a1, a0, nf, ev_op_iterations, ev_op_max_order
):
    """
    Determine used kernel and call it.

    In LO we always use the exact solution.

    Parameters
    ----------
        method : str
            method
        gamma_singlet : numpy.ndarray
            singlet anomalous dimensions matrices
        a1 : float
            target coupling value
        a0 : float
            initial coupling value
        nf : int
            number of active flavors
        ev_op_iterations : int
            number of evolution steps
        ev_op_max_order : int
            perturbative expansion order of U

    Returns
    -------
        e_s : numpy.ndarray
            singlet EKO
    """
    # use always exact in LO
    if order == 0:
        return lo_exact(gamma_singlet, a1, a0, nf)
    # NLO
    if method == "decompose-exact":
        return nlo_decompose_exact(gamma_singlet, a1, a0, nf)
    if method == "decompose-expanded":
        return nlo_decompose_expanded(gamma_singlet, a1, a0, nf)
    if method == "perturbative-exact":
        return nlo_perturbative_exact(
            gamma_singlet, a1, a0, nf, ev_op_iterations, ev_op_max_order
        )
    if method == "perturbative-expanded":
        return nlo_perturbative_expanded(
            gamma_singlet, a1, a0, nf, ev_op_iterations, ev_op_max_order
        )
    if method in ["truncated", "ordered-truncated"]:
        return nlo_truncated(gamma_singlet, a1, a0, nf, ev_op_iterations)
    # if method in ["iterate-exact", "iterate-expanded"]:
    return nlo_iterate(gamma_singlet, a1, a0, nf, ev_op_iterations)



# -*- coding: utf-8 -*-
"""
This file contains the next-to-leading-order Altarelli-Parisi splitting kernels.

These expression have been obtained using the procedure described in the
`wiki <https://github.com/N3PDF/eko/wiki/Parse-NLO-expressions>`_
involving ``FormGet`` :cite:`Hahn:2016ebn`.
"""

import numpy as np
import numba as nb

from .. import ekomath
from .. import constants


@nb.njit("c16(c16,u1)", cache=True)
def gamma_nsm_1(n, nf: int):
    """
    Computes the next-to-leading-order valence-like non-singlet anomalous dimension.

    Implements Eq. (3.5) of :cite:`Moch:2004pa`.

    Parameters
    ----------
        n : complex
            Mellin moment
        nf : int
            Number of active flavours

    Returns
    -------
        gamma_nsm_1 : complex
            Next-to-leading-order valence-like non-singlet anomalous dimension
            :math:`\\gamma_{ns,-}^{(1)}(N)`
    """
    S1 = ekomath.harmonic_S1(n)
    S2 = ekomath.harmonic_S2(n)
    # Here, Sp refers to S' ("s-prime") (german: "s-strich" or in Pegasus language: SSTR)
    # of :cite:`Gluck:1989ze` and NOT to the Spence function a.k.a. dilogarithm
    Sp1m = ekomath.harmonic_S1((n - 1) / 2)
    Sp2m = ekomath.harmonic_S2((n - 1) / 2)
    Sp3m = ekomath.harmonic_S3((n - 1) / 2)
    g3n = ekomath.mellin_g3(n)
    zeta2 = ekomath.zeta2
    zeta3 = ekomath.zeta3
    # fmt: off
    gqq1m_cfca = 16*g3n - (144 + n*(1 + n)*(156 + n*(340 + n*(655 + 51*n*(2 + n)))))/(18.*np.power(n,3)*np.power(1 + n,3)) + (-14.666666666666666 + 8/n - 8/(1 + n))*S2 - (4*Sp2m)/(n + np.power(n,2)) + S1*(29.77777777777778 + 16/np.power(n,2) - 16*S2 + 8*Sp2m) + 2*Sp3m + 10*zeta3 + zeta2*(16*S1 - 16*Sp1m - (16*(1 + n*np.log(2)))/n) # pylint: disable=line-too-long
    gqq1m_cfcf = -32*g3n + (24 - n*(-32 + 3*n*(-8 + n*(3 + n)*(3 + np.power(n,2)))))/(2.*np.power(n,3)*np.power(1 + n,3)) + (12 - 8/n + 8/(1 + n))*S2 + S1*(-24/np.power(n,2) - 8/np.power(1 + n,2) + 16*S2 - 16*Sp2m) + (8*Sp2m)/(n + np.power(n,2)) - 4*Sp3m - 20*zeta3 + zeta2*(-32*S1 + 32*Sp1m + 32*(1/n + np.log(2))) # pylint: disable=line-too-long
    gqq1m_cfnf = (-12 + n*(20 + n*(47 + 3*n*(2 + n))))/(9.*np.power(n,2)*np.power(1 + n,2)) - (40*S1)/9. + (8*S2)/3. # pylint: disable=line-too-long
    # fmt: on
    result = constants.CF * (
        (constants.CA * gqq1m_cfca)
        + (constants.CF * gqq1m_cfcf)
        + (2.0 * constants.TR * nf * gqq1m_cfnf)
    )
    return result


@nb.njit("c16(c16,u1)", cache=True)
def gamma_nsp_1(n, nf: int):
    """
    Computes the next-to-leading-order singlet-like non-singlet anomalous dimension.

    Implements Eq. (3.5) of :cite:`Moch:2004pa`.

    Parameters
    ----------
        n : complex
            Mellin moment
        nf : int
            Number of active flavours

    Returns
    -------
        gamma_nsp_1 : complex
            Next-to-leading-order singlet-like non-singlet anomalous dimension
            :math:`\\gamma_{ns,+}^{(1)}(N)`
    """
    S1 = ekomath.harmonic_S1(n)
    S2 = ekomath.harmonic_S2(n)
    Sp1p = ekomath.harmonic_S1(n / 2)
    Sp2p = ekomath.harmonic_S2(n / 2)
    Sp3p = ekomath.harmonic_S3(n / 2)
    g3n = ekomath.mellin_g3(n)
    zeta2 = ekomath.zeta2
    zeta3 = ekomath.zeta3
    # fmt: off
    gqq1p_cfca = -16*g3n + (132 - n*(340 + n*(655 + 51*n*(2 + n))))/(18.*np.power(n,2)*np.power(1 + n,2)) + (-14.666666666666666 + 8/n - 8/(1 + n))*S2 - (4*Sp2p)/(n + np.power(n,2)) + S1*(29.77777777777778 - 16/np.power(n,2) - 16*S2 + 8*Sp2p) + 2*Sp3p + 10*zeta3 + zeta2*(16*S1 - 16*Sp1p + 16*(1/n - np.log(2))) # pylint: disable=line-too-long
    gqq1p_cfcf = 32*g3n - (8 + n*(32 + n*(40 + 3*n*(3 + n)*(3 + np.power(n,2)))))/(2.*np.power(n,3)*np.power(1 + n,3)) + (12 - 8/n + 8/(1 + n))*S2 + S1*(40/np.power(n,2) - 8/np.power(1 + n,2) + 16*S2 - 16*Sp2p) + (8*Sp2p)/(n + np.power(n,2)) - 4*Sp3p - 20*zeta3 + zeta2*(-32*S1 + 32*Sp1p + 32*(-(1/n) + np.log(2))) # pylint: disable=line-too-long
    gqq1p_cfnf = (-12 + n*(20 + n*(47 + 3*n*(2 + n))))/(9.*np.power(n,2)*np.power(1 + n,2)) - (40*S1)/9. + (8*S2)/3. # pylint: disable=line-too-long
    # fmt: on
    result = constants.CF * (
        (constants.CA * gqq1p_cfca)
        + (constants.CF * gqq1p_cfcf)
        + (2.0 * constants.TR * nf * gqq1p_cfnf)
    )
    return result


@nb.njit("c16(c16,u1)", cache=True)
def gamma_ps_1(n, nf: int):
    """
    Computes the next-to-leading-order pure-singlet quark-quark anomalous dimension.

    Implements Eq. (3.6) of :cite:`Vogt:2004mw`.

    Parameters
    ----------
        n : complex
            Mellin moment
        nf : int
            Number of active flavours

    Returns
    -------
        gamma_ps_1 : complex
            Next-to-leading-order pure-singlet quark-quark anomalous dimension
            :math:`\\gamma_{ps}^{(1)}(N)`
    """
    # fmt: off
    gqqps1_nfcf = (-4*(2 + n*(5 + n))*(4 + n*(4 + n*(7 + 5*n))))/((-1 + n)*np.power(n,3)*np.power(1 + n,3)*np.power(2 + n,2)) # pylint: disable=line-too-long
    # fmt: on
    result = 2.0 * constants.TR * nf * constants.CF * gqqps1_nfcf
    return result


@nb.njit("c16(c16,u1)", cache=True)
def gamma_qg_1(n, nf: int):
    """
    Computes the next-to-leading-order quark-gluon singlet anomalous dimension.

    Implements Eq. (3.7) of :cite:`Vogt:2004mw`.

    Parameters
    ----------
        n : complex
            Mellin moment
        nf : int
            Number of active flavours

    Returns
    -------
        gamma_qg_1 : complex
            Next-to-leading-order quark-gluon singlet anomalous dimension
            :math:`\\gamma_{qg}^{(1)}(N)`
    """
    S1 = ekomath.harmonic_S1(n)
    S2 = ekomath.harmonic_S2(n)
    Sp2p = ekomath.harmonic_S2(n / 2)
    # fmt: off
    gqg1_nfca = (-4*(16 + n*(64 + n*(104 + n*(128 + n*(85 + n*(36 + n*(25 + n*(15 + n*(6 + n))))))))))/((-1 + n)*np.power(n,3)*np.power(1 + n,3)*np.power(2 + n,3)) - (16*(3 + 2*n)*S1)/np.power(2 + 3*n + np.power(n,2),2) + (4*(2 + n + np.power(n,2))*np.power(S1,2))/(n*(2 + 3*n + np.power(n,2))) - (4*(2 + n + np.power(n,2))*S2)/(n*(2 + 3*n + np.power(n,2))) + (4*(2 + n + np.power(n,2))*Sp2p)/(n*(2 + 3*n + np.power(n,2))) # pylint: disable=line-too-long
    gqg1_nfcf = (-2*(4 + n*(8 + n*(1 + n)*(25 + n*(26 + 5*n*(2 + n))))))/(np.power(n,3)*np.power(1 + n,3)*(2 + n)) + (8*S1)/np.power(n,2) - (4*(2 + n + np.power(n,2))*np.power(S1,2))/(n*(2 + 3*n + np.power(n,2))) + (4*(2 + n + np.power(n,2))*S2)/(n*(2 + 3*n + np.power(n,2))) # pylint: disable=line-too-long
    # fmt: on
    result = (
        2.0 * constants.TR * nf * (constants.CA * gqg1_nfca + constants.CF * gqg1_nfcf)
    )
    return result


@nb.njit("c16(c16,u1)", cache=True)
def gamma_gq_1(n, nf: int):
    """
    Computes the next-to-leading-order gluon-quark singlet anomalous dimension.

    Implements Eq. (3.8) of :cite:`Vogt:2004mw`.

    Parameters
    ----------
        n : complex
            Mellin moment
        nf : int
            Number of active flavours

    Returns
    -------
        gamma_gq_1 : complex
            Next-to-leading-order gluon-quark singlet anomalous dimension
            :math:`\\gamma_{gq}^{(1)}(N)`
    """
    S1 = ekomath.harmonic_S1(n)
    S2 = ekomath.harmonic_S2(n)
    Sp2p = ekomath.harmonic_S2(n / 2)
    # fmt: off
    ggq1_cfcf = (-8 + 2*n*(-12 + n*(-1 + n*(28 + n*(43 + 6*n*(5 + 2*n))))))/((-1 + n)*np.power(n,3)*np.power(1 + n,3)) - (4*(10 + n*(17 + n*(8 + 5*n)))*S1)/((-1 + n)*n*np.power(1 + n,2)) + (4*(2 + n + np.power(n,2))*np.power(S1,2))/(n*(-1 + np.power(n,2))) + (4*(2 + n + np.power(n,2))*S2)/(n*(-1 + np.power(n,2))) # pylint: disable=line-too-long
    ggq1_cfca = (-4*(144 + n*(432 + n*(-152 + n*(-1304 + n*(-1031 + n*(695 + n*(1678 + n*(1400 + n*(621 + 109*n))))))))))/(9.*np.power(n,3)*np.power(1 + n,3)*np.power(-2 + n + np.power(n,2),2)) + (4*(-12 + n*(-22 + 41*n + 17*np.power(n,3)))*S1)/(3.*np.power(-1 + n,2)*np.power(n,2)*(1 + n)) + ((8 + 4*n + 4*np.power(n,2))*np.power(S1,2))/(n - np.power(n,3)) + ((8 + 4*n + 4*np.power(n,2))*S2)/(n - np.power(n,3)) + (4*(2 + n + np.power(n,2))*Sp2p)/(n*(-1 + np.power(n,2))) # pylint: disable=line-too-long
    ggq1_cfnf = (8*(16 + n*(27 + n*(13 + 8*n))))/(9.*(-1 + n)*n*np.power(1 + n,2)) - (8*(2 + n + np.power(n,2))*S1)/(3.*n*(-1 + np.power(n,2))) # pylint: disable=line-too-long
    # fmt: on
    result = constants.CF * (
        (constants.CA * ggq1_cfca)
        + (constants.CF * ggq1_cfcf)
        + (2.0 * constants.TR * nf * ggq1_cfnf)
    )
    return result


@nb.njit("c16(c16,u1)", cache=True)
def gamma_gg_1(n, nf: int):
    """
    Computes the next-to-leading-order gluon-gluon singlet anomalous dimension.

    Implements Eq. (3.9) of :cite:`Vogt:2004mw`.

    Parameters
    ----------
        n : complex
            Mellin moment
        nf : int
            Number of active flavours

    Returns
    -------
        gamma_gg_1 : complex
            Next-to-leading-order gluon-gluon singlet anomalous dimension
            :math:`\\gamma_{gg}^{(1)}(N)`
    """
    S1 = ekomath.harmonic_S1(n)
    Sp1p = ekomath.harmonic_S1(n / 2)
    Sp2p = ekomath.harmonic_S2(n / 2)
    Sp3p = ekomath.harmonic_S3(n / 2)
    g3n = ekomath.mellin_g3(n)
    zeta2 = ekomath.zeta2
    zeta3 = ekomath.zeta3
    # fmt: off
    ggg1_caca = 16*g3n - (2*(576 + n*(1488 + n*(560 + n*(-1248 + n*(-1384 + n*(1663 + n*(4514 + n*(4744 + n*(3030 + n*(1225 + 48*n*(7 + n))))))))))))/(9.*np.power(-1 + n,2)*np.power(n,3)*np.power(1 + n,3)*np.power(2 + n,3)) + S1*(29.77777777777778 + 16/np.power(-1 + n,2) + 16/np.power(1 + n,2) - 16/np.power(2 + n,2) - 8*Sp2p) + (16*(1 + n + np.power(n,2))*Sp2p)/(n*(1 + n)*(-2 + n + np.power(n,2))) - 2*Sp3p - 10*zeta3 + zeta2*(-16*S1 + 16*Sp1p + 16*(-(1/n) + np.log(2))) # pylint: disable=line-too-long
    ggg1_canf = (8*(6 + n*(1 + n)*(28 + n*(1 + n)*(13 + 3*n*(1 + n)))))/(9.*np.power(n,2)*np.power(1 + n,2)*(-2 + n + np.power(n,2))) - (40*S1)/9. # pylint: disable=line-too-long
    ggg1_cfnf = (2*(-8 + n*(-8 + n*(-10 + n*(-22 + n*(-3 + n*(6 + n*(8 + n*(4 + n)))))))))/(np.power(n,3)*np.power(1 + n,3)*(-2 + n + np.power(n,2))) # pylint: disable=line-too-long
    # fmt: on
    result = constants.CA * constants.CA * ggg1_caca + 2.0 * constants.TR * nf * (
        constants.CA * ggg1_canf + constants.CF * ggg1_cfnf
    )
    return result


@nb.njit("c16[:,:](c16,u1)", cache=True)
def gamma_singlet_1(N, nf: int):
    r"""
      Computes the next-leading-order singlet anomalous dimension matrix

      .. math::
          \gamma_S^{(1)} = \left(\begin{array}{cc}
            \gamma_{qq}^{(1)} & \gamma_{qg}^{(1)}\\
            \gamma_{gq}^{(1)} & \gamma_{gg}^{(1)}
          \end{array}\right)

      Parameters
      ----------
        N : complex
          Mellin moment
        nf : int
          Number of active flavours

      Returns
      -------
        gamma_S_1 : numpy.ndarray
          Next-to-leading-order singlet anomalous dimension matrix :math:`\gamma_{S}^{(1)}(N)`

      See Also
      --------
        gamma_nsp_1 : :math:`\gamma_{qq}^{(1)}`
        gamma_ps_1 : :math:`\gamma_{qq}^{(1)}`
        gamma_qg_1 : :math:`\gamma_{qg}^{(1)}`
        gamma_gq_1 : :math:`\gamma_{gq}^{(1)}`
        gamma_gg_1 : :math:`\gamma_{gg}^{(1)}`
    """
    gamma_qq = gamma_nsp_1(N, nf) + gamma_ps_1(N, nf)
    gamma_qg = gamma_qg_1(N, nf)
    gamma_gq = gamma_gq_1(N, nf)
    gamma_gg = gamma_gg_1(N, nf)
    gamma_S_0 = np.array([[gamma_qq, gamma_qg], [gamma_gq, gamma_gg]], np.complex_)
    return gamma_S_0



# -*- coding: utf-8 -*-
r"""
This module contains the central operator classes.

See :doc:`Operator overview </Code/Operators>`.
"""

import time
import logging

import numpy as np
from scipy import integrate
import numba as nb

from .. import mellin
from .. import interpolation
from .. import anomalous_dimensions as ad
from .. import beta
from ..kernels import non_singlet as ns
from ..kernels import singlet as s

from .member import OpMember
from .physical import PhysicalOperator

logger = logging.getLogger(__name__)


@nb.njit("c16[:](u1,string,c16,u1,f8)", cache=True)
def gamma_ns_fact(order, mode, n, nf, L):
    gamma_ns = ad.gamma_ns(order, mode[-1], n, nf)
    if order > 0:
        gamma_ns[1] -= beta.beta(0, nf) * gamma_ns[0] * L
    return gamma_ns


@nb.njit("c16[:,:,:](u1,c16,u1,f8)", cache=True)
def gamma_singlet_fact(order, n, nf, L):
    gamma_singlet = ad.gamma_singlet(
        order,
        n,
        nf,
    )
    if order > 0:
        gamma_singlet[1] -= beta.beta(0, nf) * gamma_singlet[0] * L
    return gamma_singlet


@nb.njit("f8(f8,u1,string,string,b1,f8,f8[:,:],f8,f8,f8,f8,u4,u1)", cache=True)
def quad_ker(
    u,
    order,
    mode,
    method,
    is_log,
    logx,
    areas,
    a1,
    a0,
    nf,
    L,
    ev_op_iterations,
    ev_op_max_order,
):
    """
    Raw kernel inside quad

    Parameters
    ----------
        u : float
            quad argument
        order : int
            perturbation order
        method : str
            method
        mode : str
            element in the singlet sector
        is_log : boolean
            logarithmic interpolation
        logx : float
            Mellin inversion point
        areas : tuple
            basis function configuration
        a1 : float
            target coupling value
        a0 : float
            initial coupling value
        nf : int
            number of active flavors
        L : float
            logarithmic ratio of factorization and renormalization scale
        ev_op_iterations : int
            number of evolution steps
        ev_op_max_order : int
            perturbative expansion order of U

    Returns
    -------
        ker : float
            evaluated integration kernel
    """
    is_singlet = mode[0] == "S"
    # get transformation to N integral
    if is_singlet:
        r, o = 0.4 * 16.0 / (1.0 - logx), 1.0
    else:
        r, o = 0.5, 0.0
    n = mellin.Talbot_path(u, r, o)
    jac = mellin.Talbot_jac(u, r, o)
    # check PDF is active
    if is_log:
        pj = interpolation.log_evaluate_Nx(n, logx, areas)
    else:
        pj = interpolation.evaluate_Nx(n, logx, areas)
    if pj == 0.0:
        return 0.0
    # compute the actual evolution kernel
    if is_singlet:
        gamma_singlet = gamma_singlet_fact(order, n, nf, L)
        ker = s.dispatcher(
            order, method, gamma_singlet, a1, a0, nf, ev_op_iterations, ev_op_max_order
        )
        # select element of matrix
        k = 0 if mode[2] == "q" else 1
        l = 0 if mode[3] == "q" else 1
        ker = ker[k, l]
    else:
        # load data
        gamma_ns = gamma_ns_fact(order, mode, n, nf, L)
        # switch by order and method
        ker = ns.dispatcher(
            order,
            method,
            gamma_ns,
            a1,
            a0,
            nf,
            ev_op_iterations,
        )
    # recombine everthing
    mellin_prefactor = np.complex(0.0, -1.0 / np.pi)
    return np.real(mellin_prefactor * ker * pj * jac)


class Operator:
    """
    Internal representation of a single EKO.

    The actual matrices are computed only upon calling :meth:`compute`.
    :meth:`compose` will generate the :class:`PhysicalOperator` for the outside world.
    If not computed yet, :meth:`compose` will call :meth:`compute`.

    Parameters
    ----------
        master : eko.operator_grid.OperatorMaster
            the master instance
        q2_from : float
            evolution source
        q2_to : float
            evolution target
        mellin_cut : float
            cut to the upper limit in the mellin inversion
    """

    def __init__(self, config, managers, nf, q2_from, q2_to, mellin_cut=1e-2):
        self.config = config
        self.managers = managers
        self.nf = nf
        self.q2_from = q2_from
        self.q2_to = q2_to
        # TODO make 'cut' external parameter?
        self._mellin_cut = mellin_cut
        self.op_members = {}

    def compose(self, op_list, instruction_set, q2_final):
        """
        Compose all :class:`Operator` together.

        Calls :meth:`compute`, if necessary.

        Parameters
        ----------
            op_list : list(Operator)
                list of operators to merge
            instruction_set : dict
                list of instructions (generated by :class:`eko.thresholds.FlavourTarget`)
            q2_final : float
                final scale

        Returns
        -------
            op : PhysicalOperator
                final operator
        """
        # compute?
        if len(self.op_members.keys()) == 0:
            self.compute()
        # prepare operators
        op_to_compose = [self.op_members] + [i.op_members for i in reversed(op_list)]
        # iterate operators
        new_ops = {}
        for name, instructions in instruction_set:
            for origin, paths in instructions.items():
                key = f"{name}.{origin}"
                op = OpMember.join(op_to_compose, paths)
                # enforce new name
                op.name = key
                new_ops[key] = op
        return PhysicalOperator(new_ops, q2_final)

    def labels(self):
        """
        Compute necessary sector labels to compute.

        Returns
        -------
            labels : list(str)
                sector labels
        """
        order = self.config["order"]
        labels = []
        # NS sector is dynamic
        if self.config["debug_skip_non_singlet"]:
            logger.warning("Evolution: skipping non-singlet sector")
        else:
            labels.append("NS_p")
            if order > 0:
                labels.append("NS_m")
        # singlet sector is fixed
        if self.config["debug_skip_singlet"]:
            logger.warning("Evolution: skipping singlet sector")
        else:
            labels.extend(["S_qq", "S_qg", "S_gq", "S_gg"])
        return labels

    def compute(self):
        """ compute the actual operators (i.e. run the integrations) """
        # Generic parameters
        int_disp = self.managers["interpol_dispatcher"]
        grid_size = len(int_disp.xgrid)

        # init all ops with zeros
        labels = self.labels()
        for n in ["S_qq", "S_qg", "S_gq", "S_gg", "NS_p", "NS_m", "NS_v"]:
            self.op_members[n] = OpMember(
                np.zeros((grid_size, grid_size)), np.zeros((grid_size, grid_size)), n
            )
        tot_start_time = time.perf_counter()
        # setup KernelDispatcher
        logger.info("Evolution: computing operators - 0/%d", grid_size)
        sc = self.managers["strong_coupling"]
        fact_to_ren = self.config["fact_to_ren"]
        a1 = sc.a_s(self.q2_to / fact_to_ren, self.q2_to)
        a0 = sc.a_s(self.q2_from / fact_to_ren, self.q2_from)
        # iterate output grid
        for k, logx in enumerate(np.log(int_disp.xgrid_raw)):
            start_time = time.perf_counter()
            # iterate basis functions
            for l, bf in enumerate(int_disp):
                # iterate sectors
                for label in labels:
                    # compute and set
                    res = integrate.quad(
                        quad_ker,
                        0.5,
                        1.0 - self._mellin_cut,
                        args=(
                            self.config["order"],
                            label,
                            self.config["method"],
                            int_disp.log,
                            logx,
                            bf.areas_representation,
                            a1,
                            a0,
                            self.nf,
                            np.log(fact_to_ren),
                            self.config["ev_op_iterations"],
                            self.config["ev_op_max_order"],
                        ),
                        epsabs=1e-12, epsrel=1e-5,
                        limit=100,
                        full_output=1,
                    )
                    val,err = res[:2]
                    self.op_members[label].value[k][l] = val
                    self.op_members[label].error[k][l] = err

            logger.info(
                "Evolution: computing operators - %d/%d took: %f s",
                k + 1,
                grid_size,
                time.perf_counter() - start_time,
            )

        # closing comment
        logger.info("Evolution: Total time %f s", time.perf_counter() - tot_start_time)
        # copy non-singlet kernels, if necessary
        self.copy_ns_ops()

    def copy_ns_ops(self):
        """Copy non-singlet kernels, if necessary"""
        order = self.config["order"]
        if order == 0:  # in LO +=-=v
            for label in ["NS_v", "NS_m"]:
                self.op_members[label].value = self.op_members["NS_p"].value.copy()
                self.op_members[label].error = self.op_members["NS_p"].error.copy()
        elif order == 1:  # in NLO -=v
            self.op_members["NS_v"].value = self.op_members["NS_m"].value.copy()
            self.op_members["NS_v"].error = self.op_members["NS_m"].error.copy()



# -*- coding: utf-8 -*-
r"""
This module contains the Altarelli-Parisi splitting kernels.

Normalization is given by

.. math::
    \mathbf{P}(x) = \sum\limits_{j=0} a_s^{j+1} \mathbf P^{(j)}(x)

with :math:`a_s = \frac{\alpha_S(\mu^2)}{4\pi}`.
The 3-loop references for the non-singlet :cite:`Moch:2004pa`
and singlet :cite:`Vogt:2004mw` case contain also the lower
order results. The results are also determined in Mellin space in
terms of the anomalous dimensions (note the additional sign!)

.. math::
    \gamma(N) = - \mathcal{M}[\mathbf{P}(x)](N)
"""

import numpy as np

import numba as nb

from ..ekomath import harmonic_S1 as S1

from . import lo
from . import nlo


@nb.njit("Tuple((c16[:,:],c16,c16,c16[:,:],c16[:,:]))(c16[:,:])", cache=True)
def exp_singlet(gamma_S):
    r"""
    Computes the exponential and the eigensystem of the singlet anomalous dimension matrix

    Parameters
    ----------
        gamma_S : numpy.ndarray
            singlet anomalous dimension matrix

    Returns
    -------
        exp : numpy.ndarray
            exponential of the singlet anomalous dimension matrix :math:`\gamma_{S}(N)`
        lambda_p : complex
            positive eigenvalue of the singlet anomalous dimension matrix
            :math:`\gamma_{S}(N)`
        lambda_m : complex
            negative eigenvalue of the singlet anomalous dimension matrix
            :math:`\gamma_{S}(N)`
        e_p : numpy.ndarray
            projector for the positive eigenvalue of the singlet anomalous
            dimension matrix :math:`\gamma_{S}(N)`
        e_m : numpy.ndarray
            projector for the negative eigenvalue of the singlet anomalous
            dimension matrix :math:`\gamma_{S}(N)`

    See Also
    --------
        eko.anomalous_dimensions.lo.gamma_singlet_0 : :math:`\gamma_{S}^{(0)}(N)`
        eko.anomalous_dimensions.nlo.gamma_singlet_1 : :math:`\gamma_{S}^{(1)}(N)`
    """
    # compute eigenvalues
    det = np.sqrt(
        np.power(gamma_S[0, 0] - gamma_S[1, 1], 2) + 4.0 * gamma_S[0, 1] * gamma_S[1, 0]
    )
    lambda_p = 1.0 / 2.0 * (gamma_S[0, 0] + gamma_S[1, 1] + det)
    lambda_m = 1.0 / 2.0 * (gamma_S[0, 0] + gamma_S[1, 1] - det)
    # compute projectors
    identity = np.identity(2, np.complex_)
    c = 1.0 / det
    e_p = +c * (gamma_S - lambda_m * identity)
    e_m = -c * (gamma_S - lambda_p * identity)
    exp = e_m * np.exp(lambda_m) + e_p * np.exp(lambda_p)
    return exp, lambda_p, lambda_m, e_p, e_m


@nb.njit("c16[:](u1,string,c16,u1)", cache=True)
def gamma_ns(order, mode, n, nf):
    r"""
    Computes the tower of the non-singlet anomalous dimensions

    Parameters
    ----------
        order : int
            perturbative order
        mode : "m" | "p" | "v"
            sector identifier
        n : complex
            Mellin variable
        nf : int
            Number of active flavours

    Returns
    -------
        gamma_ns : numpy.ndarray
            non-singlet anomalous dimensions

    See Also
    --------
        eko.anomalous_dimensions.lo.gamma_ns_0 : :math:`\gamma_{ns}^{(0)}(N)`
        eko.anomalous_dimensions.nlo.gamma_nsp_1 : :math:`\gamma_{ns,+}^{(1)}(N)`
        eko.anomalous_dimensions.nlo.gamma_nsm_1 : :math:`\gamma_{ns,-}^{(1)}(N)`
    """
    # cache the s-es
    s1 = S1(n)
    gamma_ns = np.zeros(order + 1, np.complex_)
    gamma_ns[0] = lo.gamma_ns_0(n, s1)
    if order > 0:
        if mode == "p":
            gamma_ns_1 = nlo.gamma_nsp_1(n, nf)
        elif mode == "m":
            gamma_ns_1 = nlo.gamma_nsm_1(n, nf)
        gamma_ns[1] = gamma_ns_1
    return gamma_ns


@nb.njit("c16[:,:,:](u1,c16,u1)", cache=True)
def gamma_singlet(order, n, nf):
    r"""
    Computes the tower of the singlet anomalous dimensions matrices

    Parameters
    ----------
        order : int
            perturbative order
        n : complex
            Mellin variable
        nf : int
            Number of active flavours

    Returns
    -------
        gamma_singlet : numpy.ndarray
            singlet anomalous dimensions matrices

    See Also
    --------
        eko.anomalous_dimensions.lo.gamma_singlet_0 : :math:`\gamma_{S}^{(0)}(N)`
        eko.anomalous_dimensions.nlo.gamma_singlet_1 : :math:`\gamma_{S}^{(1)}(N)`
    """
    # cache the s-es
    s1 = S1(n)
    gamma_singlet = np.zeros((order + 1, 2, 2), np.complex_)
    gamma_singlet[0] = lo.gamma_singlet_0(n, s1, nf)
    if order > 0:
        gamma_singlet[1] = nlo.gamma_singlet_1(n, nf)
    return gamma_singlet



# -*- coding: utf-8 -*-
# Installation script for python
from setuptools import setup, find_packages

# write version on the fly - inspired by numpy
MAJOR = 0
MINOR = 5
MICRO = 2
ISRELEASED = True
SHORT_VERSION = "%d.%d" % (MAJOR, MINOR)
VERSION = "%d.%d.%d" % (MAJOR, MINOR, MICRO)


def write_version_py(filename="src/eko/version.py"):
    cnt = """
# THIS FILE IS GENERATED FROM SETUP.PY
major = %(major)d
short_version = '%(short_version)s'
version = '%(version)s'
full_version = '%(full_version)s'
is_released = %(isreleased)s
"""
    FULLVERSION = VERSION
    if not ISRELEASED:
        FULLVERSION += "-develop"

    a = open(filename, "w")
    try:
        a.write(
            cnt
            % {
                "major": MAJOR,
                "short_version": SHORT_VERSION,
                "version": VERSION,
                "full_version": FULLVERSION,
                "isreleased": str(ISRELEASED),
            }
        )
    finally:
        a.close()

def setup_package():
    # write version
    write_version_py()
    # paste Readme
    with open("README.md", "r") as fh:
        long_description = fh.read()
    # do it
    setup(name='eko',
        version=VERSION,
        description='Evolution Kernel Operator',
        long_description=long_description,
        long_description_content_type="text/markdown",
        author = 'A. Candido, S. Carrazza, J. Cruz-Martinez, F. Hekhorn',
        author_email='stefano.carrazza@cern.ch',
        url='https://github.com/N3PDF/eko',
        package_dir={'': 'src'},
        packages=find_packages('src'),
        package_data = {
            '' : ['doc/source/img/Logo.png'],
        },
        classifiers=[
            'Operating System :: Unix',
            'Programming Language :: Python',
            'Programming Language :: Python :: 3',
            'Topic :: Scientific/Engineering',
            'Topic :: Scientific/Engineering :: Physics',
            'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        ],
        install_requires=[
            'numpy',
            'scipy',
            'numba',
            'joblib',
            'pyyaml',
        ],
        setup_requires=[
            'wheel'
        ],
        python_requires='>=3.7'
    )

if __name__ == "__main__":
    setup_package()


# -*- coding: utf-8 -*-
r"""
This files sets the physical constants.


The constants are:

    - :math:`N_C` the number of colors - defaults to :math:`3`
    - :math:`T_R` the normalization of fundamental generators - defaults to :math:`1/2`
    - :math:`C_A` second Casimir constant in the adjoint representation - defaults to
      :math:`N_C = 3`
    - :math:`C_F` second Casimir constant in the fundamental representation - defaults to
      :math:`\frac{N_C^2-1}{2N_C} = 4/3`
"""

NC = 3
"""the number of colors"""

TR = float(1.0 / 2.0)
"""the normalization of fundamental generators"""

CA = float(NC)
"""second Casimir constant in the adjoint representation"""

CF = float((NC * NC - 1.0) / (2.0 * NC))
"""second Casimir constant in the fundamental representation"""



# -*- coding: utf-8 -*-
"""
    This file contains the main application class of eko
"""
import logging
import copy

logger = logging.getLogger(__name__)


class Runner:
    """
    Represents a single input configuration.

    For details about the configuration, see :doc:`here </Code/IO>`

    Parameters
    ----------
        setup : dict
            input configurations
    """

    def __init__(self, setup):
        # Print theory id setup
        logger.info("init Runner with %s", setup)
        # defer numba compilation
        from . import interpolation  # pylint: disable=import-outside-toplevel
        from .output import Output  # pylint: disable=import-outside-toplevel
        from .strong_coupling import (  # pylint: disable=import-outside-toplevel
            StrongCoupling,
        )
        from .thresholds import (  # pylint: disable=import-outside-toplevel
            ThresholdsConfig,
        )
        from .operator.grid import (  # pylint: disable=import-outside-toplevel
            OperatorGrid,
        )

        self.out = Output()
        if setup.get("keep_input", False):
            self.out.update(copy.deepcopy(setup))

        # setup basis grid
        bfd = interpolation.InterpolatorDispatcher.from_dict(setup)
        self.out.update(bfd.to_dict())
        # FNS
        tc = ThresholdsConfig.from_dict(setup)
        self.out["q2_ref"] = float(tc.q2_ref)
        # strong coupling
        sc = StrongCoupling.from_dict(setup, tc)
        # setup operator grid
        self.op_grid = OperatorGrid.from_dict(
            setup,
            tc,
            sc,
            bfd,
        )

    def get_operators(self):
        """compute the actual operators"""
        operators = self.op_grid.compute_q2grid()
        return operators

    def get_output(self):
        """
        Collects all data for output (to run the evolution)

        Returns
        -------
            ret : eko.output.Output
                output instance
        """
        # add all operators
        Q2grid = {}
        for op in self.get_operators():
            final_scale = op.q2_final
            Q2grid[float(final_scale)] = op.to_raw()
        self.out["Q2grid"] = Q2grid
        return copy.deepcopy(self.out)



# -*- coding: utf-8 -*-
"""
Colletion of non-singlet EKOs.
"""

import numpy as np

import numba as nb

from .. import beta

from . import evolution_integrals as ei
from . import utils


@nb.njit("c16(c16[:],f8,f8,u1)", cache=True)
def lo_exact(gamma_ns, a1, a0, nf):
    """
    Non-singlet leading order exact EKO

    Parameters
    ----------
        gamma_ns : numpy.ndarray
            non-singlet anomalous dimensions
        a1 : float
            target coupling value
        a0 : float
            initial coupling value
        nf : int
            number of active flavors

    Returns
    -------
        e_ns^0 : complex
            non-singlet leading order exact EKO
    """
    return np.exp(gamma_ns[0] * ei.j00(a1, a0, nf))


@nb.njit("c16(c16[:],f8,f8,u1)", cache=True)
def nlo_exact(gamma_ns, a1, a0, nf):
    """
    Non-singlet next-to-leading order exact EKO

    Parameters
    ----------
        gamma_ns : numpy.ndarray
            non-singlet anomalous dimensions
        a1 : float
            target coupling value
        a0 : float
            initial coupling value
        nf : int
            number of active flavors

    Returns
    -------
        e_ns^1 : complex
            non-singlet next-to-leading order exact EKO
    """
    return np.exp(
        gamma_ns[0] * ei.j01_exact(a1, a0, nf) + gamma_ns[1] * ei.j11_exact(a1, a0, nf)
    )


@nb.njit("c16(c16[:],f8,f8,u1)", cache=True)
def nlo_expanded(gamma_ns, a1, a0, nf):
    """
    Non-singlet next-to-leading order expanded EKO

    Parameters
    ----------
        gamma_ns : numpy.ndarray
            non-singlet anomalous dimensions
        a1 : float
            target coupling value
        a0 : float
            initial coupling value
        nf : int
            number of active flavors

    Returns
    -------
        e_ns^1 : complex
            non-singlet next-to-leading order expanded EKO
    """
    return np.exp(
        gamma_ns[0] * ei.j01_expanded(a1, a0, nf)
        + gamma_ns[1] * ei.j11_expanded(a1, a0, nf)
    )


@nb.njit("c16(c16[:],f8,f8,u1,u4)", cache=True)
def nlo_truncated(gamma_ns, a1, a0, nf, ev_op_iterations):
    """
    Non-singlet next-to-leading order truncated EKO

    Parameters
    ----------
        gamma_ns : numpy.ndarray
            non-singlet anomalous dimensions
        a1 : float
            target coupling value
        a0 : float
            initial coupling value
        nf : int
            number of active flavors
        ev_op_iterations : int
            number of evolution steps

    Returns
    -------
        e_ns^1 : complex
            non-singlet next-to-leading order truncated EKO
    """
    a_steps = utils.geomspace(a0, a1, ev_op_iterations)
    b1 = beta.b(1, nf)
    e = 1.0
    al = a_steps[0]
    for ah in a_steps[1:]:
        e0 = lo_exact(gamma_ns, ah, al, nf)
        e *= e0 * (1.0 + ei.j11_expanded(ah, al, nf) * (gamma_ns[1] - b1 * gamma_ns[0]))
        al = ah
    return e


@nb.njit("c16(c16[:],f8,f8,u1,u4)", cache=True)
def nlo_ordered_truncated(gamma_ns, a1, a0, nf, ev_op_iterations):
    """
    Non-singlet next-to-leading order ordered-truncated EKO

    Parameters
    ----------
        gamma_ns : numpy.ndarray
            non-singlet anomalous dimensions
        a1 : float
            target coupling value
        a0 : float
            initial coupling value
        nf : int
            number of active flavors
        ev_op_iterations : int
            number of evolution steps

    Returns
    -------
        e_ns^1 : complex
            non-singlet next-to-leading order ordered-truncated EKO
    """
    a_steps = utils.geomspace(a0, a1, ev_op_iterations)
    beta0 = beta.beta(0, nf)
    b1 = beta.b(1, nf)
    e = 1.0
    al = a_steps[0]
    for ah in a_steps[1:]:
        e0 = lo_exact(gamma_ns, ah, al, nf)
        e *= (
            e0
            * (1.0 + ah / beta0 * (gamma_ns[1] - b1 * gamma_ns[0]))
            / (1.0 + al / beta0 * (gamma_ns[1] - b1 * gamma_ns[0]))
        )
        al = ah
    return e


@nb.njit("c16(u1,string,c16[:],f8,f8,u1,u4)", cache=True)
def dispatcher(order, method, gamma_ns, a1, a0, nf, ev_op_iterations):
    """
    Determine used kernel and call it.

    In LO we always use the exact solution.

    Parameters
    ----------
        order : int
            perturbation order
        method : str
            method
        gamma_ns : numpy.ndarray
            non-singlet anomalous dimensions
        a1 : float
            target coupling value
        a0 : float
            initial coupling value
        nf : int
            number of active flavors
        ev_op_iterations : int
            number of evolution steps

    Returns
    -------
        e_ns : complex
            non-singlet EKO
    """
    # use always exact in LO
    if order == 0:
        return lo_exact(gamma_ns, a1, a0, nf)
    # NLO
    if method in ["iterate-expanded", "decompose-expanded", "perturbative-expanded"]:
        return nlo_expanded(gamma_ns, a1, a0, nf)
    elif method == "truncated":
        return nlo_truncated(gamma_ns, a1, a0, nf, ev_op_iterations)
    elif method == "ordered-truncated":
        return nlo_ordered_truncated(gamma_ns, a1, a0, nf, ev_op_iterations)
    # if method in ["iterate-exact", "decompose-exact", "perturbative-exact"]:
    return nlo_exact(gamma_ns, a1, a0, nf)



# -*- coding: utf-8 -*-
r"""
    This module contains the implementation of the
    `inverse Mellin transformation <https://en.wikipedia.org/wiki/Mellin_inversion_theorem>`_.

    It contains the actual transformations itself, as well as the necessary tools
    such as the definition of paths.

    The integral routine is provided by :func:`scipy.integrate.quad`.

    Integration Paths
    -----------------

    Although this module provides four different path implementations in practice
    only the Talbot path :cite:`Abate`

    .. math::
        p_{\text{Talbot}}(t) =  o + r \cdot ( \theta \cot(\theta) + i\theta)\quad
            \text{with}~\theta = \pi(2t-1)

    is used, as it results in the most efficient convergence. The default values
    for the parameters :math:`r,o` are given by :math:`r = 1/2, o = 0` for
    the non-singlet integrals and by :math:`r = \frac{2}{5} \frac{16}{1 - \ln(x)}, o = 1`
    for the singlet sector. Note that the non-singlet kernels evolve poles only up to
    :math:`N=0` whereas the singlet kernels have poles up to :math:`N=1`.

"""  # pylint:disable=line-too-long

import numpy as np
import numba as nb


@nb.njit("c16(f8,f8,f8)", cache=True)
def Talbot_path(t, r, o):
    """
    Talbot path.

    .. math::
        p_{\\text{Talbot}}(t) =  o + r \\cdot ( \\theta \\cot(\\theta) + i\\theta ),
        \\theta = \\pi(2t-1)

    Parameters
    ----------
        r : float
            scaling parameter - effectivly corresponds to the intersection of the path with the
            real axis
        o : float
            offset on real axis

    Returns
    -------
        path : complex
            Talbot path
    """
    theta = np.pi * (2.0 * t - 1.0)
    re = 0.0
    if t == 0.5:  # treat singular point seperately
        re = 1.0
    else:
        re = theta / np.tan(theta)
    im = theta
    return o + r * np.complex(re, im)


@nb.njit("c16(f8,f8,f8)", cache=True)
def Talbot_jac(t, r, o):  # pylint: disable=unused-argument
    """
    Derivative of Talbot path.

    .. math::
        p_{\\text{Talbot}}(t) =  o + r \\cdot ( \\theta \\cot(\\theta) + i\\theta ),
        \\theta = \\pi(2t-1)

    Parameters
    ----------
        r : float
            scaling parameter - effectivly corresponds to the intersection of the path with the
            real axis
        o : float
            offset on real axis

    Returns
    -------
        jac : complex
            derivative of Talbot path
    """
    theta = np.pi * (2.0 * t - 1.0)
    re = 0.0
    if t == 0.5:  # treat singular point seperately
        re = 0.0
    else:
        re = 1.0 / np.tan(theta)
        re -= theta / (np.sin(theta)) ** 2
    im = 1.0
    return r * np.pi * 2.0 * np.complex(re, im)


def get_path_line():
    """
    Textbook path, i.e. a straight line parallel to the imaginary axis.

    .. math::
        p_{\\text{line}}(t) = c + m \\cdot (2t - 1)

    Returns the path and its derivative which then have to be called with the arguments
    listed under `Other Parameters`.

    Other Parameters
    ----------------
        m : float
            half length of the path
        c : float
            intersection of path with real axis

    Returns
    -------
        path : function
            textbook path
        jac : function
            derivative of textbook path
    """

    @nb.njit
    def path(t, m, c):
        return np.complex(c, m * (2 * t - 1))

    @nb.njit
    def jac(_t, m, _c):
        return np.complex(0, m * 2)

    return path, jac


def get_path_edge():
    """
    Edged path with a given angle.

    .. math::
        p_{\\text{edge}}(t) = c + m\\left|t - \\frac 1 2\\right|\\exp(i\\phi)

    Returns the path and its derivative which then have to be called with the arguments
    listed under `Other Parameters`.

    Other Parameters
    ----------------
        m : float
            length of the path
        c : float, optional
            intersection of path with real axis - defaults to 1
        phi : complex, optional
            bended angle - defaults to +135° with respect to positive x axis
    Returns
    -------
        path : function
            Edged path
        jac : function
            derivative of edged path
    """

    @nb.njit
    def path(t, m, c, phi):
        if t < 0.5:  # turning point: path is not differentiable in this point
            return c + (0.5 - t) * m * np.exp(np.complex(0, -phi))
        else:
            return c + (t - 0.5) * m * np.exp(np.complex(0, +phi))

    @nb.njit
    def jac(t, m, _c, phi):
        if t < 0.5:  # turning point: jacobian is not continuous here
            return -m * np.exp(np.complex(0, -phi))
        else:
            return +m * np.exp(np.complex(0, phi))

    return path, jac



# -*- coding: utf-8 -*-
r"""
This file contains the QCD beta function coefficients and the handling of the running
coupling :math:`\alpha_s`.

See :doc:`pQCD ingredients </Theory/pQCD>`.
"""

import logging

import numpy as np
import scipy
import numba as nb

from . import constants
from . import thresholds
from .beta import beta

logger = logging.getLogger(__name__)


@nb.njit("f8(u1,f8,u1,f8,f8)", cache=True)
def as_expanded(order, as_ref, nf, scale_from, scale_to):
    """
    Compute expanded expression.

    Parameters
    ----------
        order: int
            perturbation order
        as_ref: float
            reference alpha_s
        nf: int
            value of nf for computing alpha_s
        scale_from: float
            reference scale
        scale_to : float
            target scale

    Returns
    -------
        a_s : float
            coupling at target scale :math:`a_s(Q^2)`
    """
    # common vars
    beta0 = beta(0, nf)
    lmu = np.log(scale_to / scale_from)
    den = 1.0 + beta0 * as_ref * lmu
    # LO
    as_LO = as_ref / den
    res = as_LO
    # NLO
    if order >= 1:
        beta1 = beta(1, nf)
        b1 = beta1 / beta0
        as_NLO = as_LO * (1 - b1 * as_LO * np.log(den))
        res = as_NLO
        # NNLO
        if order == 2:
            beta2 = beta(2, nf)
            b2 = beta2 / beta0
            res = as_LO * (
                1.0
                + as_LO * (as_LO - as_ref) * (b2 - b1 ** 2)
                + as_NLO * b1 * np.log(as_NLO / as_ref)
            )

    return res


class StrongCoupling:
    r"""
        Computes the strong coupling constant :math:`a_s`.

        Note that

        - all scale parameters (``scale_ref`` and ``scale_to``),
          have to be given as squared values, i.e. in units of :math:`\text{GeV}^2`
        - although, we only provide methods for
          :math:`a_s = \frac{\alpha_s(\mu^2)}{4\pi}` the reference value has to be
          given in terms of :math:`\alpha_s(\mu_0^2)` due to legacy reasons
        - the ``order`` refers to the perturbative order of the beta function, thus
          ``order=0`` means leading order beta function, means evolution with :math:`\beta_0`,
          means running at 1-loop - so there is a natural mismatch between ``order`` and the
          number of loops by one unit

        Normalization is given by :cite:`Herzog:2017ohr`:

        .. math::
            \frac{da_s(\mu^2)}{d\ln\mu^2} = \beta(a_s) \
            = - \sum\limits_{n=0} \beta_n a_s^{n+2}(\mu^2) \quad
            \text{with}~ a_s = \frac{\alpha_s(\mu^2)}{4\pi}

        See :doc:`pQCD ingredients </Theory/pQCD>`.

        Parameters
        ----------
            alpha_s_ref : float
                alpha_s(!) at the reference scale :math:`\alpha_s(\mu_0^2)`
            scale_ref : float
                reference scale :math:`\mu_0^2`
            threshold_holder : eko.thresholds.ThresholdsConfig
                An instance of :class:`~eko.thresholds.ThresholdsConfig`
            order: int
                Evaluated order of the beta function: ``0`` = LO, ...
            method : ["expanded", "exact"]
                Applied method to solve the beta function

        Examples
        --------
            >>> alpha_ref = 0.35
            >>> scale_ref = 2
            >>> threshold_holder = ThresholdsConfig( ... )
            >>> sc = StrongCoupling(alpha_ref, scale_ref, threshold_holder)
            >>> q2 = 91.1**2
            >>> sc.a_s(q2)
            0.118
    """

    def __init__(
        self,
        alpha_s_ref,
        scale_ref,
        thresh,
        order=0,
        method="exact",
    ):
        # Sanity checks
        if alpha_s_ref <= 0:
            raise ValueError(f"alpha_s_ref has to be positive - got {alpha_s_ref}")
        if scale_ref <= 0:
            raise ValueError(f"scale_ref has to be positive - got {scale_ref}")
        if not isinstance(thresh, thresholds.ThresholdsConfig):
            raise ValueError("Needs a Threshold instance")
        if order not in [0, 1, 2]:
            raise NotImplementedError("a_s beyond NNLO is not implemented")
        self._order = order
        if method not in ["expanded", "exact"]:
            raise ValueError(f"Unknown method {method}")
        self._method = method

        # create new threshold object
        self.as_ref = alpha_s_ref / 4.0 / np.pi  # convert to a_s
        if thresh.scheme == "FFNS":
            self._threshold_holder = thresholds.ThresholdsConfig(
                scale_ref, thresh.scheme, nf=thresh.nf_ref
            )
        else:
            self._threshold_holder = thresholds.ThresholdsConfig(
                scale_ref, thresh.scheme, threshold_list=thresh._area_walls
            )
        logger.info(
            "Strong Coupling: Reference a_s(Q^2=%f)=%f", self.q2_ref, self.as_ref
        )

    @property
    def q2_ref(self):
        """ reference scale """
        return self._threshold_holder.q2_ref

    @classmethod
    def from_dict(cls, setup, thresholds_config=None):
        """
        Create object from theory dictionary.

        Read keys:

            - alphas : required, reference value for  alpha_s (!)
            - Qref : required, reference value in GeV (!)
            - PTO : required, perturbative order
            - ModEv : optional, method to solve RGE, default=EXA

        Parameters
        ----------
            setup : dict
                theory dictionary
            thresholds_config : eko.thresholds.ThresholdsConfig
                threshold configuration

        Returns
        -------
            cls : StrongCoupling
                created object
        """
        # read my values
        # TODO cast to a_s here
        alpha_ref = setup["alphas"]
        q2_alpha = pow(setup["Qref"], 2)
        order = setup["PTO"]
        mod_ev = setup.get("ModEv", "EXA")
        if mod_ev in ["EXA", "iterate-exact", "decompose-exact", "perturbative-exact"]:
            method = "exact"
        elif mod_ev in [
            "TRN",
            "truncated",
            "ordered-truncated",
            "EXP",
            "iterate-expanded",
            "decompose-expanded",
            "perturbative-expanded",
        ]:
            method = "expanded"
        else:
            raise ValueError(f"Unknown evolution mode {mod_ev}")
        # eventually read my dependents
        if thresholds_config is None:
            thresholds_config = thresholds.ThresholdsConfig.from_dict(setup)
        return cls(alpha_ref, q2_alpha, thresholds_config, order, method)

    def _compute_exact(self, as_ref, nf, scale_from, scale_to):
        """
        Compute via RGE.

        Parameters
        ----------
            as_ref: float
                reference alpha_s
            nf: int
                value of nf for computing alpha_s
            scale_from: float
                reference scale
            scale_to : float
                target scale

        Returns
        -------
            a_s : float
                strong coupling at target scale :math:`a_s(Q^2)`
        """
        # in LO fallback to expanded, as this is the full solution
        if self._order == 0:
            return as_expanded(self._order, as_ref, nf, scale_from, scale_to)
        # otherwise rescale the RGE to run in terms of
        # u = beta0 * ln(scale_to/scale_from)
        beta0 = beta(0, nf)
        u = beta0 * np.log(scale_to / scale_from)
        b_vec = [1]
        # NLO
        if self._order >= 1:
            beta1 = beta(1, nf)
            b1 = beta1 / beta0
            b_vec.append(b1)
            # NNLO
            if self._order >= 2:
                beta2 = beta(2, nf)
                b2 = beta2 / beta0
                b_vec.append(b2)
        # integration kernel
        def rge(_t, a, b_vec):
            return -(a ** 2) * np.sum([a ** k * b for k, b in enumerate(b_vec)])

        # let scipy solve
        res = scipy.integrate.solve_ivp(rge, (0, u), (as_ref,), args=[b_vec])
        return res.y[0][-1]

    def _compute(self, as_ref, nf, scale_from, scale_to):
        """
        Wrapper in order to pass the computation to the corresponding
        method (depending on the calculation method).

        Parameters
        ----------
            as_ref: float
                reference alpha_s
            nf: int
                value of nf for computing alpha_s
            scale_from: float
                reference scale
            scale_to : float
                target scale

        Returns
        -------
            a_s : float
                strong coupling at target scale :math:`a_s(Q^2)`
        """
        # TODO set up a cache system here
        # at the moment everything is expanded - and type has been checked in the constructor
        if self._method == "exact":
            as_new = self._compute_exact(as_ref, nf, scale_from, scale_to)
        else:
            as_new = as_expanded(self._order, as_ref, nf, scale_from, scale_to)
        return as_new

    def a_s(self, scale_to, fact_scale=None):
        r"""
        Computes strong coupling :math:`a_s(\mu_R^2) = \frac{\alpha_s(\mu_R^2)}{4\pi}`.

        Parameters
        ----------
            scale_to : float
                final scale to evolve to :math:`\mu_R^2`

        Returns
        -------
            a_s : float
                strong coupling :math:`a_s(\mu_R^2) = \frac{\alpha_s(\mu_R^2)}{4\pi}`
        """
        # Set up the path to follow in order to go from q2_0 to q2_ref
        final_as = self.as_ref
        area_path = self._threshold_holder.get_path_from_q2_ref(scale_to)
        # as a default assume mu_F^2 = mu_R^2
        if fact_scale is None:
            fact_scale = scale_to
        for k, area in enumerate(area_path):
            q2_from = area.q2_ref
            q2_to = area.q2_towards(scale_to)
            if np.isclose(q2_from, q2_to):
                continue
            new_as = self._compute(final_as, area.nf, q2_from, q2_to)
            # apply matching conditions: see hep-ph/9706430
            # - if there is yet a step to go
            if k < len(area_path) - 1:
                next_nf_is_down = area_path[k + 1].nf < area.nf
                # q2_to is the threshold value
                L = np.log(scale_to / fact_scale)
                if next_nf_is_down:
                    c1 = -4.0 / 3.0 * constants.TR * L
                    # TODO recover color constants
                    c2 = 4.0 / 9.0 * L ** 2 - 38.0 / 3.0 * L - 14.0 / 3.0
                else:
                    c1 = 4.0 / 3.0 * constants.TR * L
                    c2 = 4.0 / 9.0 * L ** 2 + 38.0 / 3.0 * L + 14.0 / 3.0
                # shift
                if self._order == 1:
                    new_as *= 1 + c1 * new_as
                elif self._order == 2:
                    new_as *= 1 + c1 * new_as + c2 * new_as ** 2
            final_as = new_as
        return final_as
